""" FileManager (add doc here)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import six

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.FileManager.FileManagerBase import FileManagerBase
from DIRAC.Core.Utilities.List import stringListToString, intListToString, breakListIntoChunks
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.Utilities import getIDSelectString

DEBUG = 0


class FileManager(FileManagerBase):

    ######################################################
    #
    # The all important _findFiles and _getDirectoryFiles methods
    #

    def _findFiles(self, lfns, metadata=["FileID"], allStatus=False, connection=False):
        """Find file ID if it exists for the given list of LFNs"""

        connection = self._getConnection(connection)
        dirDict = self._getFileDirectories(lfns)
        failed = {}
        result = self.db.dtree.findDirs(list(dirDict))
        if not result["OK"]:
            return result
        directoryIDs = result["Value"]

        for dirPath in dirDict:
            if dirPath not in directoryIDs:
                for fileName in dirDict[dirPath]:
                    fname = "%s/%s" % (dirPath, fileName)
                    fname = fname.replace("//", "/")
                    failed[fname] = "No such file or directory"

        successful = {}
        for dirPath in directoryIDs:
            fileNames = dirDict[dirPath]
            res = self._getDirectoryFiles(
                directoryIDs[dirPath], fileNames, metadata, allStatus=allStatus, connection=connection
            )
            if (not res["OK"]) or (not res["Value"]):
                error = res.get("Message", "No such file or directory")
                for fileName in fileNames:
                    fname = "%s/%s" % (dirPath, fileName)
                    fname = fname.replace("//", "/")
                    failed[fname] = error
            else:
                for fileName, fileDict in res["Value"].items():
                    fname = "%s/%s" % (dirPath, fileName)
                    fname = fname.replace("//", "/")
                    successful[fname] = fileDict
            for fileName in fileNames:
                if fileName not in res["Value"]:
                    fname = "%s/%s" % (dirPath, fileName)
                    fname = fname.replace("//", "/")
                    failed[fname] = "No such file or directory"
        return S_OK({"Successful": successful, "Failed": failed})

    def _findFileIDs(self, lfns, connection=False):
        """Find lfn <-> FileID correspondence"""
        connection = self._getConnection(connection)
        dirDict = self._getFileDirectories(lfns)
        failed = {}
        successful = {}
        result = self.db.dtree.findDirs(list(dirDict))
        if not result["OK"]:
            return result
        directoryIDs = result["Value"]
        directoryPaths = {}

        for dirPath in dirDict:
            if dirPath not in directoryIDs:
                for fileName in dirDict[dirPath]:
                    fname = "%s/%s" % (dirPath, fileName)
                    fname = fname.replace("//", "/")
                    failed[fname] = "No such file or directory"
            else:
                directoryPaths[directoryIDs[dirPath]] = dirPath

        for dirIDs in breakListIntoChunks(list(directoryIDs), 1000):
            wheres = []
            for dirPath in dirIDs:
                fileNames = dirDict[dirPath]
                dirID = directoryIDs[dirPath]
                wheres.append("( DirID=%d AND FileName IN (%s) )" % (dirID, stringListToString(fileNames)))

            req = "SELECT FileName,DirID,FileID FROM FC_Files WHERE %s" % " OR ".join(wheres)
            result = self.db._query(req, connection)
            if not result["OK"]:
                return result
            for fileName, dirID, fileID in result["Value"]:
                fname = "%s/%s" % (directoryPaths[dirID], fileName)
                fname = fname.replace("//", "/")
                successful[fname] = fileID

        for lfn in lfns:
            if lfn not in successful:
                failed[lfn] = "No such file or directory"

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryFiles(self, dirID, fileNames, metadata_input, allStatus=False, connection=False):
        """Get the metadata for files in the same directory"""
        metadata = list(metadata_input)

        connection = self._getConnection(connection)
        # metadata can be any of ['FileID','Size','UID','GID','Status','Checksum','ChecksumType',
        # 'Type','CreationDate','ModificationDate','Mode']
        req = "SELECT FileName,DirID,FileID,Size,UID,GID,Status FROM FC_Files WHERE DirID=%d" % (dirID)
        if not allStatus:
            statusIDs = []
            for status in self.db.visibleFileStatus:
                res = self._getStatusInt(status, connection=connection)
                if res["OK"]:
                    statusIDs.append(res["Value"])
            if statusIDs:
                req = "%s AND Status IN (%s)" % (req, intListToString(statusIDs))
        if fileNames:
            req = "%s AND FileName IN (%s)" % (req, stringListToString(fileNames))
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        fileNameIDs = res["Value"]
        if not fileNameIDs:
            return S_OK({})
        filesDict = {}
        # If we only requested the FileIDs then there is no need to do anything else
        if metadata == ["FileID"]:
            for fileName, dirID, fileID, size, uid, gid, status in fileNameIDs:
                filesDict[fileName] = {"FileID": fileID}
            return S_OK(filesDict)
        # Otherwise get the additionally requested metadata from the FC_FileInfo table
        files = {}
        userDict = {}
        groupDict = {}
        for fileName, dirID, fileID, size, uid, gid, status in fileNameIDs:
            filesDict[fileID] = fileName
            files[fileName] = {}
            if "Size" in metadata:
                files[fileName]["Size"] = size
            if "DirID" in metadata:
                files[fileName]["DirID"] = dirID
            if "UID" in metadata:
                files[fileName]["UID"] = uid
                if uid in userDict:
                    owner = userDict[uid]
                else:
                    owner = "unknown"
                    result = self.db.ugManager.getUserName(uid)
                    if result["OK"]:
                        owner = result["Value"]
                    userDict[uid] = owner
                files[fileName]["Owner"] = owner
            if "GID" in metadata:
                files[fileName]["GID"] = gid
                if gid in groupDict:
                    group = groupDict[gid]
                else:
                    group = "unknown"
                    result = self.db.ugManager.getGroupName(gid)
                    if result["OK"]:
                        group = result["Value"]
                    groupDict[gid] = group
                files[fileName]["OwnerGroup"] = group
            if "Status" in metadata:
                files[fileName]["Status"] = self._getIntStatus(status).get("Value", status)
        for element in ["FileID", "Size", "DirID", "UID", "GID", "Status"]:
            if element in metadata:
                metadata.remove(element)
        metadata.append("FileID")
        metadata.reverse()
        req = "SELECT %s FROM FC_FileInfo WHERE FileID IN (%s)" % (
            intListToString(metadata),
            intListToString(filesDict.keys()),
        )
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        for tuple_ in res["Value"]:
            fileID = tuple_[0]
            rowDict = dict(zip(metadata, tuple_))
            files[filesDict[fileID]].update(rowDict)
        return S_OK(files)

    def _getDirectoryFileIDs(self, dirID, requestString=False):
        """Get a list of IDs for all the files stored in given directories or their
            subdirectories
        :param mixt dirID: single directory ID or a list of directory IDs
        :param boolean requestString: if True return result as a SQL SELECT string
        :return: list of file IDs or SELECT string
        """

        result = getIDSelectString(dirID)
        if not result["OK"]:
            return result
        dirListString = result["Value"]

        if requestString:
            req = "SELECT FileID FROM FC_Files WHERE DirID IN ( %s )" % dirListString
            return S_OK(req)

        req = "SELECT FileID,DirID,FileName FROM FC_Files WHERE DirID IN ( %s )" % dirListString
        result = self.db._query(req)
        return result

    def _getFileMetadataByID(self, fileIDs, connection=False):
        """Get standard file metadata for a list of files specified by FileID"""
        stringIDs = ",".join(["%s" % id_ for id_ in fileIDs])
        req = "SELECT FileID,Size,UID,GID,Status FROM FC_Files WHERE FileID in ( %s )" % stringIDs
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        resultDict = {}
        for fileID, size, uid, gid, status in result["Value"]:
            resultDict[fileID] = {
                "Size": int(size),
                "UID": int(uid),
                "GID": int(gid),
                "Status": self._getIntStatus(status).get("Value", status),
            }

        req = "SELECT FileID,GUID,CreationDate from FC_FileInfo WHERE FileID in ( %s )" % stringIDs
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        for fileID, guid, date in result["Value"]:
            resultDict.setdefault(fileID, {})
            resultDict[fileID].update({"GUID": guid, "CreationDate": date})

        return S_OK(resultDict)

    ######################################################
    #
    # _addFiles related methods
    #

    def _insertFiles(self, lfns, uid, gid, connection=False):

        connection = self._getConnection(connection)
        # Add the files
        failed = {}
        insertTuples = []
        res = self._getStatusInt("AprioriGood", connection=connection)
        statusID = 0
        if res["OK"]:
            statusID = res["Value"]

        directorySESizeDict = {}
        for lfn in lfns.keys():
            dirID = lfns[lfn]["DirID"]
            fileName = os.path.basename(lfn)
            size = lfns[lfn]["Size"]
            ownerDict = lfns[lfn].get("Owner", None)
            s_uid = uid
            s_gid = gid
            if ownerDict:
                result = self.db.ugManager.getUserAndGroupID(ownerDict)
                if result["OK"]:
                    s_uid, s_gid = result["Value"]
            insertTuples.append("(%d,%d,%d,%d,%d,'%s')" % (dirID, size, s_uid, s_gid, statusID, fileName))
            directorySESizeDict.setdefault(dirID, {})
            directorySESizeDict[dirID].setdefault(0, {"Files": 0, "Size": 0})
            directorySESizeDict[dirID][0]["Size"] += lfns[lfn]["Size"]
            directorySESizeDict[dirID][0]["Files"] += 1

        req = "INSERT INTO FC_Files (DirID,Size,UID,GID,Status,FileName) VALUES %s" % (",".join(insertTuples))
        res = self.db._update(req, connection)
        if not res["OK"]:
            return res
        # Get the fileIDs for the inserted files
        res = self._findFiles(list(lfns), ["FileID"], connection=connection)
        if not res["OK"]:
            for lfn in list(lfns):
                failed[lfn] = "Failed post insert check"
                lfns.pop(lfn)
        else:
            failed.update(res["Value"]["Failed"])
            for lfn in res["Value"]["Failed"]:
                lfns.pop(lfn)
            for lfn, fileDict in res["Value"]["Successful"].items():
                lfns[lfn]["FileID"] = fileDict["FileID"]
        insertTuples = []
        toDelete = []
        for lfn in lfns:
            fileInfo = lfns[lfn]
            fileID = fileInfo["FileID"]
            dirID = fileInfo["DirID"]
            checksum = fileInfo["Checksum"]
            checksumtype = fileInfo.get("ChecksumType", "Adler32")
            guid = fileInfo.get("GUID", "")
            mode = fileInfo.get("Mode", self.db.umask)
            toDelete.append(fileID)
            insertTuples.append(
                "(%d,'%s','%s','%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),%d)" % (fileID, guid, checksum, checksumtype, mode)
            )
        if insertTuples:
            fields = "FileID,GUID,Checksum,ChecksumType,CreationDate,ModificationDate,Mode"
            req = "INSERT INTO FC_FileInfo (%s) VALUES %s" % (fields, ",".join(insertTuples))
            res = self.db._update(req)
            if not res["OK"]:
                self._deleteFiles(toDelete, connection=connection)
                for lfn in list(lfns):
                    failed[lfn] = res["Message"]
                    lfns.pop(lfn)
            else:
                # Update the directory usage
                result = self._updateDirectoryUsage(directorySESizeDict, "+", connection=connection)
                if not result["OK"]:
                    gLogger.warn("Failed to insert FC_DirectoryUsage", result["Message"])

        return S_OK({"Successful": lfns, "Failed": failed})

    def _getFileIDFromGUID(self, guid, connection=False):
        connection = self._getConnection(connection)
        if not guid:
            return S_OK({})
        if not isinstance(guid, (list, tuple)):
            guid = [guid]
        req = "SELECT FileID,GUID FROM FC_FileInfo WHERE GUID IN (%s)" % stringListToString(guid)
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        guidDict = {}
        for fileID, guid in res["Value"]:
            guidDict[guid] = fileID
        return S_OK(guidDict)

    def getLFNForGUID(self, guids, connection=False):
        """Returns the lfns matching given guids"""

        connection = self._getConnection(connection)
        if not guids:
            return S_OK({})
        if not isinstance(guids, (list, tuple)):
            guids = [guids]
        req = "SELECT f.FileID, f.FileName, fi.GUID, f.DirID FROM FC_FileInfo fi"
        req += " JOIN FC_Files f on fi.FileID = f.FileID WHERE GUID IN (%s)" % stringListToString(guids)
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res

        fileIDguid = {}
        dirIDFileIDs = {}
        fileIDName = {}
        for fileID, fileName, guid, dirID in res["Value"]:
            fileIDguid[fileID] = guid
            dirIDFileIDs.setdefault(dirID, []).append(fileID)
            fileIDName[fileID] = fileName

        res = self.db.dtree.getDirectoryPaths(list(dirIDFileIDs))
        if not res["OK"]:
            return res

        dirIDName = res["Value"]

        guidLFN = dict(
            [
                (fileIDguid[fId], os.path.join(dirIDName[dId], fileIDName[fId]))
                for dId in dirIDName
                for fId in dirIDFileIDs[dId]
            ]
        )
        failedGuid = set(guids) - set(guidLFN)
        failed = dict.fromkeys(failedGuid, "GUID does not exist") if failedGuid else {}

        return S_OK({"Successful": guidLFN, "Failed": failed})

    ######################################################
    #
    # _deleteFiles related methods
    #

    def _deleteFiles(self, fileIDs, connection=False):
        connection = self._getConnection(connection)
        replicaPurge = self.__deleteFileReplicas(fileIDs)
        filePurge = self.__deleteFiles(fileIDs, connection=connection)
        if not replicaPurge["OK"]:
            return replicaPurge
        if not filePurge["OK"]:
            return filePurge
        return S_OK()

    def __deleteFileReplicas(self, fileIDs, connection=False):
        connection = self._getConnection(connection)
        res = self.__getFileIDReplicas(fileIDs, connection=connection)
        if not res["OK"]:
            return res
        return self.__deleteReplicas(list(res["Value"]), connection=connection)

    def __deleteFiles(self, fileIDs, connection=False):
        connection = self._getConnection(connection)
        if not isinstance(fileIDs, (list, tuple)):
            fileIDs = [fileIDs]
        if not fileIDs:
            return S_OK()
        fileIDString = intListToString(fileIDs)
        failed = []
        for table in ["FC_Files", "FC_FileInfo"]:
            req = "DELETE FROM %s WHERE FileID in (%s)" % (table, fileIDString)
            res = self.db._update(req, connection)
            if not res["OK"]:
                gLogger.error("Failed to remove files from table %s" % table, res["Message"])
                failed.append(table)
        if failed:
            return S_ERROR("Failed to remove files from %s" % stringListToString(failed))
        return S_OK()

    ######################################################
    #
    # _addReplicas related methods
    #

    def _insertReplicas(self, lfns, master=False, connection=False):
        connection = self._getConnection(connection)
        # Add the files
        failed = {}
        successful = {}
        insertTuples = []
        fileIDLFNs = {}
        res = self._getStatusInt("AprioriGood")
        statusID = 0
        if res["OK"]:
            statusID = res["Value"]
        for lfn in list(lfns):
            fileID = lfns[lfn]["FileID"]
            fileIDLFNs[fileID] = lfn
            seName = lfns[lfn]["SE"]
            if isinstance(seName, six.string_types):
                seList = [seName]
            elif isinstance(seName, list):
                seList = seName
            else:
                return S_ERROR("Illegal type of SE list: %s" % str(type(seName)))
            for seName in seList:
                res = self.db.seManager.findSE(seName)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                    lfns.pop(lfn)
                    continue
                seID = res["Value"]
                insertTuples.append((fileID, seID))
        if not master:
            res = self._getRepIDsForReplica(insertTuples, connection=connection)
            if not res["OK"]:
                return res
            for fileID, repDict in res["Value"].items():
                for seID, repID in repDict.items():
                    successful[fileIDLFNs[fileID]] = True
                    insertTuples.remove((fileID, seID))

        if not insertTuples:
            return S_OK({"Successful": successful, "Failed": failed})

        req = "INSERT INTO FC_Replicas (FileID,SEID,Status) VALUES %s" % (
            ",".join(["(%d,%d,%d)" % (tuple_[0], tuple_[1], statusID) for tuple_ in insertTuples])
        )
        res = self.db._update(req, connection)
        if not res["OK"]:
            return res
        res = self._getRepIDsForReplica(insertTuples, connection=connection)
        if not res["OK"]:
            return res
        replicaDict = res["Value"]
        directorySESizeDict = {}
        for fileID, repDict in replicaDict.items():
            lfn = fileIDLFNs[fileID]
            dirID = lfns[lfn]["DirID"]
            directorySESizeDict.setdefault(dirID, {})
            for seID, repID in repDict.items():
                lfns[lfn]["RepID"] = repID
                directorySESizeDict[dirID].setdefault(seID, {"Files": 0, "Size": 0})
                directorySESizeDict[dirID][seID]["Size"] += lfns[lfn]["Size"]
                directorySESizeDict[dirID][seID]["Files"] += 1

        replicaType = "Replica"
        if master:
            replicaType = "Master"
        insertReplicas = []
        toDelete = []
        for lfn in lfns.keys():
            fileDict = lfns[lfn]
            repID = fileDict.get("RepID", 0)
            if repID:
                pfn = fileDict["PFN"]
                toDelete.append(repID)
                insertReplicas.append("(%d,'%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),'%s')" % (repID, replicaType, pfn))
        if insertReplicas:
            req = "INSERT INTO FC_ReplicaInfo (RepID,RepType,CreationDate,ModificationDate,PFN) VALUES %s" % (
                ",".join(insertReplicas)
            )
            res = self.db._update(req, connection)
            if not res["OK"]:
                for lfn in lfns.keys():
                    failed[lfn] = res["Message"]
                self.__deleteReplicas(toDelete, connection=connection)
            else:
                # Update the directory usage
                self._updateDirectoryUsage(directorySESizeDict, "+", connection=connection)
                for lfn in lfns.keys():
                    successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def _getRepIDsForReplica(self, replicaTuples, connection=False):
        connection = self._getConnection(connection)
        queryTuples = []
        for fileID, seID in replicaTuples:
            queryTuples.append("(%d,%d)" % (fileID, seID))
        req = "SELECT RepID,FileID,SEID FROM FC_Replicas WHERE (FileID,SEID) IN (%s)" % intListToString(queryTuples)
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        replicaDict = {}
        for repID, fileID, seID in res["Value"]:
            replicaDict.setdefault(fileID, {})
            replicaDict[fileID][seID] = repID

        return S_OK(replicaDict)

    ######################################################
    #
    # _deleteReplicas related methods
    #

    def _deleteReplicas(self, lfns, connection=False):
        connection = self._getConnection(connection)
        failed = {}
        successful = {}
        res = self._findFiles(list(lfns), ["DirID", "FileID", "Size"], connection=connection)

        # If the file does not exist we consider the deletion successful
        for lfn, error in res["Value"]["Failed"].items():
            if error == "No such file or directory":
                successful[lfn] = True
            else:
                failed[lfn] = error

        lfnFileIDDict = res["Value"]["Successful"]
        toRemove = []
        directorySESizeDict = {}
        for lfn, fileDict in lfnFileIDDict.items():
            fileID = fileDict["FileID"]
            se = lfns[lfn]["SE"]
            if isinstance(se, six.string_types):
                res = self.db.seManager.findSE(se)
                if not res["OK"]:
                    return res
            seID = res["Value"]
            toRemove.append((fileID, seID))
            # Now prepare the storage usage update
            dirID = fileDict["DirID"]
            directorySESizeDict.setdefault(dirID, {})
            directorySESizeDict[dirID].setdefault(seID, {"Files": 0, "Size": 0})
            directorySESizeDict[dirID][seID]["Size"] += fileDict["Size"]
            directorySESizeDict[dirID][seID]["Files"] += 1
        res = self._getRepIDsForReplica(toRemove, connection)
        if not res["OK"]:
            for lfn in lfnFileIDDict.keys():
                failed[lfn] = res["Message"]
        else:
            repIDs = []
            for fileID, seDict in res["Value"].items():
                for seID, repID in seDict.items():
                    repIDs.append(repID)
            res = self.__deleteReplicas(repIDs, connection=connection)
            if not res["OK"]:
                for lfn in lfnFileIDDict.keys():
                    failed[lfn] = res["Message"]
            else:
                # Update the directory usage
                self._updateDirectoryUsage(directorySESizeDict, "-", connection=connection)
                for lfn in lfnFileIDDict.keys():
                    successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def __deleteReplicas(self, repIDs, connection=False):
        connection = self._getConnection(connection)
        if not isinstance(repIDs, (list, tuple)):
            repIDs = [repIDs]
        if not repIDs:
            return S_OK()
        repIDString = intListToString(repIDs)
        failed = []
        for table in ["FC_Replicas", "FC_ReplicaInfo"]:
            req = "DELETE FROM %s WHERE RepID in (%s)" % (table, repIDString)
            res = self.db._update(req, connection)
            if not res["OK"]:
                gLogger.error("Failed to remove replicas from table %s" % table, res["Message"])
                failed.append(table)
        if failed:
            return S_ERROR("Failed to remove replicas from %s" % stringListToString(failed))
        return S_OK()

    ######################################################
    #
    # _setReplicaStatus _setReplicaHost _setReplicaParameter methods
    # _setFileParameter method
    #

    def _setReplicaStatus(self, fileID, se, status, connection=False):
        if status not in self.db.validReplicaStatus:
            return S_ERROR("Invalid replica status %s" % status)
        connection = self._getConnection(connection)
        res = self._getStatusInt(status, connection=connection)
        if not res["OK"]:
            return res
        statusID = res["Value"]
        res = self.__getRepIDForReplica(fileID, se, connection=connection)
        if not res["OK"]:
            return res
        if not res["Exists"]:
            return S_ERROR("Replica does not exist")
        if not res["Value"]:
            return res
        repID = res["Value"]
        req = "UPDATE FC_Replicas SET Status=%d WHERE RepID=%d" % (statusID, repID)
        return self.db._update(req, connection)

    def _setReplicaHost(self, fileID, se, newSE, connection=False):
        connection = self._getConnection(connection)
        res = self.db.seManager.findSE(newSE)
        if not res["OK"]:
            return res
        newSE = res["Value"]
        res = self.__getRepIDForReplica(fileID, se, connection=connection)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return res
        repID = res["Value"]
        req = "UPDATE FC_Replicas SET SEID=%d WHERE RepID = %d;" % (newSE, repID)
        return self.db._update(req, connection)

    def _setReplicaParameter(self, fileID, se, paramName, paramValue, connection=False):
        connection = self._getConnection(connection)
        res = self.__getRepIDForReplica(fileID, se, connection=connection)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return res
        repID = res["Value"]
        req = "UPDATE FC_ReplicaInfo SET %s='%s', ModificationDate = UTC_TIMESTAMP() WHERE RepID IN (%d)" % (
            paramName,
            paramValue,
            repID,
        )
        return self.db._update(req, connection)

    def _setFileParameter(self, fileID, paramName, paramValue, connection=False):
        connection = self._getConnection(connection)

        result = getIDSelectString(fileID)
        if not result["OK"]:
            return result
        fileIDString = result["Value"]

        if paramName in ["UID", "GID", "Status", "Size"]:
            # Treat primary file attributes specially
            # Different statement for the fileIDString with SELECT is for performance optimization
            # since in this case the MySQL engine manages to use index on FileID.
            if "select" in fileIDString.lower():
                tmpreq = "UPDATE FC_Files as FF1, ( %s ) as FF2 %%s WHERE FF1.FileID=FF2.FileID" % fileIDString
            else:
                tmpreq = "UPDATE FC_Files %%s WHERE FileID IN (%s)" % fileIDString
            req = tmpreq % "SET %s='%s'" % (paramName, paramValue)
            result = self.db._update(req, connection)
            if not result["OK"]:
                return result
            if "select" in fileIDString.lower():
                req = "UPDATE FC_FileInfo as FF1, ( %s ) as FF2" % fileIDString
                req += " SET ModificationDate=UTC_TIMESTAMP() WHERE FF1.FileID=FF2.FileID"
            else:
                req = "UPDATE FC_FileInfo SET ModificationDate=UTC_TIMESTAMP() WHERE FileID IN (%s)" % fileIDString
        else:
            # Different statement for the fileIDString with SELECT is for performance optimization
            # since in this case the MySQL engine manages to use index on FileID.
            if "select" in fileIDString.lower():
                req = "UPDATE FC_FileInfo as FF1, ( %s ) as FF2 SET %s='%s', " % (fileIDString, paramName, paramValue)
                req += "ModificationDate=UTC_TIMESTAMP() WHERE FF1.FileID=FF2.FileID"
            else:
                req = "UPDATE FC_FileInfo SET %s='%s', ModificationDate=UTC_TIMESTAMP() WHERE FileID IN (%s)" % (
                    paramName,
                    paramValue,
                    fileIDString,
                )
        return self.db._update(req, connection)

    def __getRepIDForReplica(self, fileID, seID, connection=False):
        connection = self._getConnection(connection)
        if isinstance(seID, six.string_types):
            res = self.db.seManager.findSE(seID)
            if not res["OK"]:
                return res
            seID = res["Value"]
        res = self._getRepIDsForReplica([(fileID, seID)], connection=connection)
        if not res["OK"]:
            return res
        if not res["Value"]:
            result = S_OK()
            result["Exists"] = False
        else:
            result = S_OK(res["Value"][fileID][seID])
            result["Exists"] = True
        return result

    ######################################################
    #
    # _getFileReplicas related methods
    #

    def _getFileReplicas(self, fileIDs, fields_input=["PFN"], allStatus=False, connection=False):
        """Get replicas for the given list of files specified by their fileIDs"""
        fields = list(fields_input)
        connection = self._getConnection(connection)
        res = self.__getFileIDReplicas(fileIDs, allStatus=allStatus, connection=connection)
        if not res["OK"]:
            return res
        fileIDDict = res["Value"]
        if fileIDDict:
            if "Status" in fields:
                fields.remove("Status")
            repIDDict = {}
            if fields:
                req = "SELECT RepID,%s FROM FC_ReplicaInfo WHERE RepID IN (%s);" % (
                    intListToString(fields),
                    intListToString(fileIDDict.keys()),
                )
                res = self.db._query(req, connection)
                if not res["OK"]:
                    return res
                for tuple_ in res["Value"]:
                    repID = tuple_[0]
                    repIDDict[repID] = dict(zip(fields, tuple_[1:]))
                    statusID = fileIDDict[repID][2]
                    res = self._getIntStatus(statusID, connection=connection)
                    if not res["OK"]:
                        continue
                    repIDDict[repID]["Status"] = res["Value"]
            else:
                for repID in fileIDDict:
                    statusID = fileIDDict[repID][2]
                    res = self._getIntStatus(statusID, connection=connection)
                    if not res["OK"]:
                        continue
                    repIDDict[repID] = {"Status": res["Value"]}
        seDict = {}
        replicas = {}
        for repID in fileIDDict.keys():
            fileID, seID, statusID = fileIDDict[repID]
            replicas.setdefault(fileID, {})
            if seID not in seDict:
                res = self.db.seManager.getSEName(seID)
                if not res["OK"]:
                    continue
                seDict[seID] = res["Value"]
            seName = seDict[seID]
            replicas[fileID][seName] = repIDDict.get(repID, {})

        if len(replicas) != len(fileIDs):
            for fileID in fileIDs:
                if fileID not in replicas:
                    replicas[fileID] = {}

        return S_OK(replicas)

    def __getFileIDReplicas(self, fileIDs, allStatus=False, connection=False):
        connection = self._getConnection(connection)
        if not fileIDs:
            return S_ERROR("No such file or directory")
        req = "SELECT FileID,SEID,RepID,Status FROM FC_Replicas WHERE FileID IN (%s)" % (intListToString(fileIDs))
        if not allStatus:
            statusIDs = []
            for status in self.db.visibleReplicaStatus:
                result = self._getStatusInt(status, connection=connection)
                if result["OK"]:
                    statusIDs.append(result["Value"])
            req += " AND Status in (%s)" % (intListToString(statusIDs))
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        fileIDDict = {}
        for fileID, seID, repID, statusID in res["Value"]:
            fileIDDict[repID] = (fileID, seID, statusID)
        return S_OK(fileIDDict)

    def _getDirectoryReplicas(self, dirID, allStatus=False, connection=False):
        """Get replicas for files in a given directory"""
        replicaStatusIDs = []
        if not allStatus:
            for status in self.db.visibleReplicaStatus:
                result = self._getStatusInt(status, connection=connection)
                if result["OK"]:
                    replicaStatusIDs.append(result["Value"])
        fileStatusIDs = []
        if not allStatus:
            for status in self.db.visibleFileStatus:
                result = self._getStatusInt(status, connection=connection)
                if result["OK"]:
                    fileStatusIDs.append(result["Value"])

        if not self.db.lfnPfnConvention or self.db.lfnPfnConvention == "Weak":
            req = "SELECT FF.FileName,FR.FileID,FR.SEID,FI.PFN FROM FC_Files as FF,"
            req += " FC_Replicas as FR, FC_ReplicaInfo as FI"
            req += " WHERE FF.FileID=FR.FileID AND FR.RepID=FI.RepID AND FF.DirID=%d " % dirID
            if replicaStatusIDs:
                req += " AND FR.Status in (%s)" % intListToString(replicaStatusIDs)
            if fileStatusIDs:
                req += " AND FF.Status in (%s)" % intListToString(fileStatusIDs)
        else:
            req = "SELECT FF.FileName,FR.FileID,FR.SEID,'' FROM FC_Files as FF,"
            req += " FC_Replicas as FR"
            req += " WHERE FF.FileID=FR.FileID AND FF.DirID=%d " % dirID
            if replicaStatusIDs:
                req += " AND FR.Status in (%s)" % intListToString(replicaStatusIDs)
            if fileStatusIDs:
                req += " AND FF.Status in (%s)" % intListToString(fileStatusIDs)

        result = self.db._query(req, connection)
        return result

    def repairFileTables(self, connection=False):
        """Repair FC_FileInfo table by adding missing records as compaired to the FC_Files table"""

        req = "SELECT F1.FileID, F2.FileID from FC_Files as F1 LEFT JOIN FC_FileInfo as F2 "
        req += "ON F1.FileID=F2.FileID WHERE F2.FileID IS NULL"
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result

        fileIDsToAdd = []
        for f1, _ in result["Value"]:
            fileIDsToAdd.append(f1)

        if not fileIDsToAdd:
            return S_OK(0)

        nFiles = len(fileIDsToAdd)
        insertTuples = []
        for fileID in fileIDsToAdd:
            guid = makeGuid()
            insertTuples.append("(%d,'%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),%d)" % (int(fileID), guid, self.db.umask))

        fields = "FileID,GUID,CreationDate,ModificationDate,Mode"
        req = "INSERT INTO FC_FileInfo (%s) VALUES %s" % (fields, ",".join(insertTuples))
        result = self.db._update(req)
        if not result["OK"]:
            return result

        return S_OK(nFiles)



""" JobRunningMatchedRatioPolicy

  Policy that calculates the efficiency following the formula::

    ( running ) / ( running + matched + received + checking )

  if the denominator is smaller than 10, it does not take any decision.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase

__RCSID__ = "$Id$"


class JobRunningMatchedRatioPolicy(PolicyBase):
    """
    The JobRunningMatchedRatioPolicy class is a policy that checks the efficiency of the
    jobs according to what is on JobDB.

      Evaluates the JobRunningMatchedRatioPolicy results given by the JobCommand.JobCommand
    """

    @staticmethod
    def _evaluate(commandResult):
        """_evaluate

        efficiency < 0.5 :: Banned
        efficiency < 0.9 :: Degraded

        """

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        commandResult = commandResult[0]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        running = commandResult["Running"]
        matched = commandResult["Matched"]
        received = commandResult["Received"]
        checking = commandResult["Checking"]

        total = running + matched + received + checking

        # we want a minimum amount of jobs to take a decision ( at least 10 pilots )
        if total < 10:
            result["Status"] = "Unknown"
            result["Reason"] = "Not enough jobs to take a decision"
            return S_OK(result)

        efficiency = running / total

        if efficiency <= 0.5:
            result["Status"] = "Banned"
        elif efficiency <= 0.9:
            result["Status"] = "Degraded"
        else:
            result["Status"] = "Active"

        result["Reason"] = "Job Running / Matched ratio of %.2f" % efficiency
        return S_OK(result)



#####################################################################
# File: ReqManagerHandler.py
########################################################################
"""Implementation of the RequestDB service in the DISET framework


.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN ReqManager
  :end-before: ##END
  :dedent: 2
  :caption: ReqManager options


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# # imports
import six
import json
import datetime
import math

# # from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning

# # from RMS
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.RequestManagementSystem.DB.RequestDB import RequestDB


class ReqManagerHandlerMixin(object):
    """
    .. class:: ReqManagerHandler

    RequestDB interface in the DISET framework.
    """

    # # request validator
    __validator = None
    # # request DB instance
    __requestDB = None

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """initialize handler"""

        try:
            cls.__requestDB = RequestDB()
        except RuntimeError as error:
            gLogger.exception(error)
            return S_ERROR(error)

        # If there is a constant delay to be applied to each request
        cls.constantRequestDelay = getServiceOption(serviceInfoDict, "ConstantRequestDelay", 0)

        # # create tables for empty db
        return cls.__requestDB.createTables()

    # # helper functions
    @classmethod
    def validate(cls, request):
        """request validation"""
        if not cls.__validator:
            cls.__validator = RequestValidator()
        return cls.__validator.validate(request)

    types_getRequestIDForName = [six.string_types]

    @classmethod
    def export_getRequestIDForName(cls, requestName):
        """get requestID for given :requestName:"""
        if isinstance(requestName, six.string_types):
            result = cls.__requestDB.getRequestIDForName(requestName)
            if not result["OK"]:
                return result
            requestID = result["Value"]
        return S_OK(requestID)

    types_cancelRequest = [six.integer_types]

    @classmethod
    def export_cancelRequest(cls, requestID):
        """Cancel a request"""
        return cls.__requestDB.cancelRequest(requestID)

    types_putRequest = [six.string_types]

    def export_putRequest(self, requestJSON):
        """put a new request into RequestDB

        :param cls: class ref
        :param str requestJSON: request serialized to JSON format
        """
        requestDict = json.loads(requestJSON)
        requestName = requestDict.get("RequestID", requestDict.get("RequestName", "***UNKNOWN***"))
        request = Request(requestDict)

        # Check whether the credentials in the Requests are correct and allowed to be set
        isAuthorized = RequestValidator.setAndCheckRequestOwner(request, self.getRemoteCredentials())

        if not isAuthorized:
            return S_ERROR(DErrno.ENOAUTH, "Credentials in the requests are not allowed")

        optimized = request.optimize()
        if optimized.get("Value", False):
            gLogger.debug("putRequest: request was optimized")
        else:
            gLogger.debug("putRequest: request unchanged", optimized.get("Message", "Nothing could be optimized"))

        valid = self.validate(request)
        if not valid["OK"]:
            gLogger.error("putRequest: request %s not valid: %s" % (requestName, valid["Message"]))
            return valid

        # If NotBefore is not set or user defined, we calculate its value

        now = datetime.datetime.utcnow().replace(microsecond=0)
        extraDelay = datetime.timedelta(0)
        if request.Status not in Request.FINAL_STATES and (not request.NotBefore or request.NotBefore < now):
            # We don't delay if it is the first insertion
            if getattr(request, "RequestID", 0):
                # If it is a constant delay, just set it
                if self.constantRequestDelay:
                    extraDelay = datetime.timedelta(minutes=self.constantRequestDelay)
                else:
                    # If there is a waiting Operation with Files
                    op = request.getWaiting().get("Value")
                    if op and len(op):
                        attemptList = [opFile.Attempt for opFile in op if opFile.Status == "Waiting"]
                        if attemptList:
                            maxWaitingAttempt = max([opFile.Attempt for opFile in op if opFile.Status == "Waiting"])
                            # In case it is the first attempt, extraDelay is 0
                            # maxWaitingAttempt can be None if the operation has no File, like the ForwardDiset
                            extraDelay = datetime.timedelta(
                                minutes=2 * math.log(maxWaitingAttempt) if maxWaitingAttempt else 0
                            )

                request.NotBefore = now + extraDelay

        gLogger.info(
            "putRequest: request %s not before %s (extra delay %s)"
            % (request.RequestName, request.NotBefore, extraDelay)
        )

        requestName = request.RequestName
        gLogger.info("putRequest: Attempting to set request '%s'" % requestName)
        return self.__requestDB.putRequest(request)

    types_getScheduledRequest = [six.integer_types]

    @classmethod
    def export_getScheduledRequest(cls, operationID):
        """read scheduled request given operationID"""
        scheduled = cls.__requestDB.getScheduledRequest(operationID)
        if not scheduled["OK"]:
            gLogger.error("getScheduledRequest: %s" % scheduled["Message"])
            return scheduled
        if not scheduled["Value"]:
            return S_OK()
        requestJSON = scheduled["Value"].toJSON()
        if not requestJSON["OK"]:
            gLogger.error("getScheduledRequest: %s" % requestJSON["Message"])
        return requestJSON

    types_getDBSummary = []

    @classmethod
    def export_getDBSummary(cls):
        """Get the summary of requests in the Request DB"""
        return cls.__requestDB.getDBSummary()

    types_getRequest = [six.integer_types]

    @classmethod
    def export_getRequest(cls, requestID=0):
        """Get a request of given type from the database"""
        getRequest = cls.__requestDB.getRequest(requestID)
        if not getRequest["OK"]:
            gLogger.error("getRequest: %s" % getRequest["Message"])
            return getRequest
        if getRequest["Value"]:
            getRequest = getRequest["Value"]
            toJSON = getRequest.toJSON()
            if not toJSON["OK"]:
                gLogger.error(toJSON["Message"])
            return toJSON
        return S_OK()

    types_getBulkRequests = [int, bool]

    @classmethod
    @ignoreEncodeWarning
    def export_getBulkRequests(cls, numberOfRequest, assigned):
        """Get a request of given type from the database

        :warning: the dictionary may contain string keys instead of int (json serialization)
                  Do not forget to cast it back

        :param numberOfRequest: size of the bulk (default 10)
        :return: S_OK( {Failed : message, Successful : list of Request.toJSON()} )
        """
        getRequests = cls.__requestDB.getBulkRequests(numberOfRequest=numberOfRequest, assigned=assigned)
        if not getRequests["OK"]:
            gLogger.error("getRequests: %s" % getRequests["Message"])
            return getRequests
        if getRequests["Value"]:
            getRequests = getRequests["Value"]
            toJSONDict = {"Successful": {}, "Failed": {}}

            for rId in getRequests:
                toJSON = getRequests[rId].toJSON()
                if not toJSON["OK"]:
                    gLogger.error(toJSON["Message"])
                    toJSONDict["Failed"][rId] = toJSON["Message"]
                else:
                    toJSONDict["Successful"][rId] = toJSON["Value"]
            return S_OK(toJSONDict)
        return S_OK()

    types_peekRequest = [six.integer_types]

    @classmethod
    def export_peekRequest(cls, requestID=0):
        """peek request given its id"""
        peekRequest = cls.__requestDB.peekRequest(requestID)
        if not peekRequest["OK"]:
            gLogger.error("peekRequest: %s" % peekRequest["Message"])
            return peekRequest
        if peekRequest["Value"]:
            peekRequest = peekRequest["Value"].toJSON()
            if not peekRequest["OK"]:
                gLogger.error(peekRequest["Message"])
        return peekRequest

    types_getRequestSummaryWeb = [dict, list, int, int]

    @classmethod
    def export_getRequestSummaryWeb(cls, selectDict, sortList, startItem, maxItems):
        """Returns a list of Request for the web portal

        :param dict selectDict: parameter on which to restrain the query {key : Value}
                                key can be any of the Request columns, 'Type' (interpreted as Operation.Type)
                                and 'FromData' and 'ToData' are matched against the LastUpdate field
        :param sortList: [sorting column, ASC/DESC]
        :type sortList: python:list
        :param int startItem: start item (for pagination)
        :param int maxItems: max items (for pagination)
        """
        return cls.__requestDB.getRequestSummaryWeb(selectDict, sortList, startItem, maxItems)

    types_getDistinctValuesWeb = [six.string_types]

    @classmethod
    def export_getDistinctValuesWeb(cls, attribute):
        """Get distinct values for a given request attribute. 'Type' is interpreted as
        the operation type"""

        tableName = "Request"
        if attribute == "Type":
            tableName = "Operation"
        return cls.__requestDB.getDistinctValues(tableName, attribute)

    types_getRequestCountersWeb = [six.string_types, dict]

    @classmethod
    def export_getRequestCountersWeb(cls, groupingAttribute, selectDict):
        """For the web portal.
        Returns a dictionary {value : counts} for a given key.
        The key can be any field from the RequestTable. or "Type",
        which will be interpreted as 'Operation.Type'

        :param groupingAttribute: attribute used for grouping
        :param selectDict: selection criteria
        """

        return cls.__requestDB.getRequestCountersWeb(groupingAttribute, selectDict)

    types_deleteRequest = [six.integer_types]

    @classmethod
    def export_deleteRequest(cls, requestID):
        """Delete the request with the supplied ID"""
        return cls.__requestDB.deleteRequest(requestID)

    types_getRequestIDsList = [list, int, six.string_types]

    @classmethod
    def export_getRequestIDsList(cls, statusList=None, limit=None, since=None, until=None, getJobID=False):
        """get requests' IDs with status in :statusList:"""
        statusList = statusList if statusList else list(Request.FINAL_STATES)
        limit = limit if limit else 100
        since = since if since else ""
        until = until if until else ""
        reqIDsList = cls.__requestDB.getRequestIDsList(statusList, limit, since=since, until=until, getJobID=getJobID)
        if not reqIDsList["OK"]:
            gLogger.error("getRequestIDsList: %s" % reqIDsList["Message"])
        return reqIDsList

    types_getRequestIDsForJobs = [list]

    @classmethod
    def export_getRequestIDsForJobs(cls, jobIDs):
        """Select the request IDs for supplied jobIDs

        :warning: the dictionary may contain string keys instead of int (json serialization)
          Do not forget to cast it back

        """
        return cls.__requestDB.getRequestIDsForJobs(jobIDs)

    types_readRequestsForJobs = [list]

    @classmethod
    @ignoreEncodeWarning
    def export_readRequestsForJobs(cls, jobIDs):
        """read requests for jobs given list of jobIDs"""
        requests = cls.__requestDB.readRequestsForJobs(jobIDs)
        if not requests["OK"]:
            gLogger.error("readRequestsForJobs: %s" % requests["Message"])
            return requests
        for jobID, request in requests["Value"]["Successful"].items():
            requests["Value"]["Successful"][jobID] = request.toJSON()["Value"]
        return requests

    types_getDigest = [six.integer_types]

    @classmethod
    def export_getDigest(cls, requestID):
        """get digest for a request given its id

        :param str requestID: request's id
        :return: S_OK( json_str )
        """
        return cls.__requestDB.getDigest(requestID)

    types_getRequestStatus = [six.integer_types]

    @classmethod
    def export_getRequestStatus(cls, requestID):
        """get request status given its id"""
        status = cls.__requestDB.getRequestStatus(requestID)
        if not status["OK"]:
            gLogger.error("getRequestStatus: %s" % status["Message"])
        return status

    types_getRequestFileStatus = [list(six.integer_types), list(six.string_types) + [list]]

    @classmethod
    def export_getRequestFileStatus(cls, requestID, lfnList):
        """get request file status for a given LFNs list and requestID"""
        if isinstance(lfnList, six.string_types):
            lfnList = [lfnList]
        res = cls.__requestDB.getRequestFileStatus(requestID, lfnList)
        if not res["OK"]:
            gLogger.error("getRequestFileStatus: %s" % res["Message"])
        return res

    types_getRequestInfo = [list(six.integer_types)]

    @classmethod
    def export_getRequestInfo(cls, requestID):
        """get request info for a given requestID"""
        requestInfo = cls.__requestDB.getRequestInfo(requestID)
        if not requestInfo["OK"]:
            gLogger.error("getRequestInfo: %s" % requestInfo["Message"])
        return requestInfo


class ReqManagerHandler(ReqManagerHandlerMixin, RequestHandler):
    pass



""" DIRAC FileCatalog Security Manager base class
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Security.Properties import FC_MANAGEMENT

_readMethods = [
    "exists",
    "isFile",
    "getFileSize",
    "getFileMetadata",
    "getReplicas",
    "getReplicaStatus",
    "getFileAncestors",
    "getFileDescendents",
    "listDirectory",
    "isDirectory",
    "getDirectoryReplicas",
    "getDirectorySize",
    "getDirectoryMetadata",
]

_writeMethods = [
    "changePathOwner",
    "changePathGroup",
    "changePathMode",
    "addFile",
    "setFileStatus",
    "removeFile",
    "addReplica",
    "removeReplica",
    "setReplicaStatus",
    "setReplicaHost",
    "addFileAncestors",
    "createDirectory",
    "removeDirectory",
    "setMetadata",
    "__removeMetadata",
]


class SecurityManagerBase(object):
    def __init__(self, database=None):
        self.db = database

    def setDatabase(self, database):
        self.db = database

    def getPathPermissions(self, paths, credDict):
        """Get path permissions according to the policy"""
        return S_ERROR("The getPathPermissions method must be implemented in the inheriting class")

    def hasAccess(self, opType, paths, credDict):
        # Map the method name to Read/Write
        if opType in _readMethods:
            opType = "Read"
        elif opType in _writeMethods:
            opType = "Write"

        # Check if admin access is granted first
        result = self.hasAdminAccess(credDict)
        if not result["OK"]:
            return result
        if result["Value"]:
            # We are admins, allow everything
            permissions = {}
            for path in paths:
                permissions[path] = True
            return S_OK({"Successful": permissions, "Failed": {}})

        successful = {}
        failed = {}
        if not opType.lower() in ["read", "write", "execute"]:
            return S_ERROR("Operation type not known")
        if self.db.globalReadAccess and (opType.lower() == "read"):
            for path in paths:
                successful[path] = True
            resDict = {"Successful": successful, "Failed": {}}
            return S_OK(resDict)

        result = self.getPathPermissions(paths, credDict)
        if not result["OK"]:
            return result

        permissions = result["Value"]["Successful"]
        for path, permDict in permissions.items():
            if permDict[opType]:
                successful[path] = True
            else:
                successful[path] = False

        failed.update(result["Value"]["Failed"])

        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    def hasAdminAccess(self, credDict):
        if FC_MANAGEMENT in credDict["properties"]:
            return S_OK(True)
        return S_OK(False)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities.ThreadSafe import Synchronizer
from DIRAC.FrameworkSystem.Client.Logger import gLogger

gEventSync = Synchronizer()


class EventDispatcher(object):
    def __init__(self):
        self.__events = {}
        self.__processingEvents = set()

    @gEventSync
    def registerEvent(self, eventName):
        if eventName in self.__events:
            return
        self.__events[eventName] = []

    @gEventSync
    def addListener(self, eventName, functor):
        if eventName not in self.__events:
            return S_ERROR("Event %s is not registered" % eventName)
        if functor in self.__events[eventName]:
            return S_OK()
        self.__events[eventName].append(functor)
        return S_OK()

    @gEventSync
    def removeListener(self, eventName, functor):
        if eventName not in self.__events:
            return S_ERROR("Event %s is not registered" % eventName)
        if functor not in self.__events[eventName]:
            return S_OK()
        iPos = self.__events[eventName].find(functor)
        del self.__events[eventName][iPos]
        return S_OK()

    def isEventBeingProcessed(self, eventName):
        return eventName in self.__processingEvents

    def getRegisteredEvents(self):
        return sorted(self.__events)

    def triggerEvent(self, eventName, params=False, threaded=False):
        if threaded:
            th = threading.Thread(target=self.__realTrigger, args=(eventName, params))
            th.setDaemon(1)
            th.start()
            return S_OK(0)
        return self.__realTrigger(eventName, params)

    def __realTrigger(self, eventName, params):
        gEventSync.lock()
        try:
            if eventName not in self.__events:
                return S_ERROR("Event %s is not registered" % eventName)
            if eventName in self.__processingEvents:
                return S_OK(0)
            eventFunctors = list(self.__events[eventName])
            self.__processingEvents.add(eventName)
        finally:
            gEventSync.unlock()
        finalResult = S_OK()
        for functor in eventFunctors:
            try:
                result = functor(eventName, params)
            except Exception:
                gLogger.exception("Listener %s for event %s raised an exception" % (functor.__name__, eventName))
                continue
            if not isinstance(result, dict) or "OK" not in result:
                gLogger.error(
                    "Listener for event did not return a S_OK/S_ERROR structure",
                    "%s %s" % (functor.__name__, eventName),
                )
                continue
            if not result["OK"]:
                finalResult = result
                break
        gEventSync.lock()
        try:
            self.__processingEvents.discard(eventName)
        finally:
            try:
                gEventSync.unlock()
            except Exception:
                pass
        if not finalResult["OK"]:
            return finalResult
        return S_OK(len(eventFunctors))


gEventDispatcher = EventDispatcher()



########################################################################
# File: FileCatalogHandler.py
########################################################################
"""
:mod: FileCatalogHandler

.. module: FileCatalogHandler

:synopsis: FileCatalogHandler is a simple Replica and Metadata Catalog service

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# imports
import csv

from six import StringIO

# from DIRAC

from DIRAC import gLogger, S_ERROR
from DIRAC.DataManagementSystem.Service.FileCatalogHandler import FileCatalogHandlerMixin

from DIRAC.Core.Tornado.Server.TornadoService import TornadoService


sLog = gLogger.getSubLogger(__name__)


class TornadoFileCatalogHandler(FileCatalogHandlerMixin, TornadoService):
    """
    ..class:: FileCatalogHandler

    A simple Replica and Metadata Catalog service.
    """

    # This is needed because the mixin class uses `cls.log`
    log = sLog

    def export_streamToClient(self, seName):
        """This method used to transfer the SEDump to the client,
        formated as CSV with '|' separation

        :param seName: name of the se to dump

        :returns: the result of the FileHelper


        """

        retVal = self.getSEDump(seName)

        try:
            csvOutput = StringIO()
            writer = csv.writer(csvOutput, delimiter="|")
            writer.writerows(retVal)

            ret = csvOutput.getvalue()
            return ret

        except Exception as e:
            sLog.exception("Exception while sending seDump", repr(e))
            return S_ERROR("Exception while sendind seDump: %s" % repr(e))
        finally:
            csvOutput.close()



#! /usr/bin/env python
########################################################################
# File :    dirac-admin-set-site-protocols
# Author :  Stuart Paterson
########################################################################
"""
Defined protocols for each SE for a given site.

Example:
  $ dirac-admin-set-site-protocols --Site=LCG.IN2P3.fr SRM2
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("", "Site=", "Site for which protocols are to be set (mandatory)")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["Protocol: SE access protocol"], mandatory=False)
    switches, args = Script.parseCommandLine(ignoreErrors=True)

    site = None
    for switch in switches:
        if switch[0].lower() == "site":
            site = switch[1]

    if not site or not args:
        Script.showHelp(exitCode=1)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    result = diracAdmin.setSiteProtocols(site, args, printOutput=True)
    if not result["OK"]:
        print("ERROR: %s" % result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" CommandCaller

  Module that loads commands and executes them.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import copy

from DIRAC import S_ERROR, S_OK
from DIRAC.ResourceStatusSystem.Utilities import Utils


def commandInvocation(commandTuple, pArgs=None, decisionParams=None, clients=None):
    """
    Returns a command object, given commandTuple

    :params:
      `commandTuple`: a tuple, where commandTuple[0] is a module name and
      commandTuple[1] is a class name (inside the module)
    """

    if commandTuple is None:
        return S_OK(None)

    # decision params can be a dictionary passed with all the element parameters
    # used mostly by the PDP to inject all relevant information
    if decisionParams is None:
        decisionParams = {}

    # arguments hardcoded on Configurations.py for the policy
    if pArgs is None:
        pArgs = {}

    try:
        cModule = commandTuple[0]
        cClass = commandTuple[1]
        commandModule = Utils.voimport("DIRAC.ResourceStatusSystem.Command." + cModule)
    except ImportError:
        return S_ERROR("Import error for command %s." % (cModule))

    if not hasattr(commandModule, cClass):
        return S_ERROR("%s has no %s" % (cModule, cClass))

    # We merge decision parameters and policy arguments.
    newArgs = copy.deepcopy(decisionParams)
    newArgs.update(pArgs)

    commandObject = getattr(commandModule, cClass)(newArgs, clients)

    return S_OK(commandObject)


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" Module that holds DISET Authorization class for services
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Security import Properties
from DIRAC.Core.Utilities import List

__RCSID__ = "$Id$"


class AuthManager(object):
    """Handle Service Authorization"""

    __authLogger = gLogger.getSubLogger("Authorization")
    KW_HOSTS_GROUP = "hosts"
    KW_DN = "DN"
    KW_GROUP = "group"
    KW_EXTRA_CREDENTIALS = "extraCredentials"
    KW_PROPERTIES = "properties"
    KW_USERNAME = "username"

    def __init__(self, authSection):
        """
        Constructor

        :type authSection: string
        :param authSection: Section containing the authorization rules
        """
        self.authSection = authSection

    def authQuery(self, methodQuery, credDict, defaultProperties=False):
        """
        Check if the query is authorized for a credentials dictionary

        :type  methodQuery: string
        :param methodQuery: Method to test
        :type  credDict: dictionary
        :param credDict: dictionary containing credentials for test. The dictionary can contain the DN
                            and selected group.
        :return: Boolean result of test
        """
        userString = ""
        if self.KW_DN in credDict:
            userString += "DN=%s" % credDict[self.KW_DN]
        if self.KW_GROUP in credDict:
            userString += " group=%s" % credDict[self.KW_GROUP]
        if self.KW_EXTRA_CREDENTIALS in credDict:
            userString += " extraCredentials=%s" % str(credDict[self.KW_EXTRA_CREDENTIALS])
        self.__authLogger.debug("Trying to authenticate %s" % userString)
        # Get properties
        requiredProperties = self.getValidPropertiesForMethod(methodQuery, defaultProperties)
        # Extract valid groups
        validGroups = self.getValidGroups(requiredProperties)
        lowerCaseProperties = [prop.lower() for prop in requiredProperties]
        if not lowerCaseProperties:
            lowerCaseProperties = ["any"]

        allowAll = "any" in lowerCaseProperties or "all" in lowerCaseProperties
        # Set no properties by default
        credDict[self.KW_PROPERTIES] = []
        # Check non secure backends
        if self.KW_DN not in credDict or not credDict[self.KW_DN]:
            if allowAll and not validGroups:
                self.__authLogger.debug("Accepted request from unsecure transport")
                return True
            else:
                self.__authLogger.debug(
                    "Explicit property required and query seems to be coming through an unsecure transport"
                )
                return False
        # Check if query comes though a gateway/web server
        if self.forwardedCredentials(credDict):
            self.__authLogger.debug("Query comes from a gateway")
            self.unpackForwardedCredentials(credDict)
            return self.authQuery(methodQuery, credDict, requiredProperties)
        # Get the properties
        # Check for invalid forwarding
        if self.KW_EXTRA_CREDENTIALS in credDict:
            # Invalid forwarding?
            if not isinstance(credDict[self.KW_EXTRA_CREDENTIALS], six.string_types):
                self.__authLogger.debug("The credentials seem to be forwarded by a host, but it is not a trusted one")
                return False
        # Is it a host?
        if self.KW_EXTRA_CREDENTIALS in credDict and credDict[self.KW_EXTRA_CREDENTIALS] == self.KW_HOSTS_GROUP:
            # Get the nickname of the host
            credDict[self.KW_GROUP] = credDict[self.KW_EXTRA_CREDENTIALS]
        # HACK TO MAINTAIN COMPATIBILITY
        else:
            if self.KW_EXTRA_CREDENTIALS in credDict and self.KW_GROUP not in credDict:
                credDict[self.KW_GROUP] = credDict[self.KW_EXTRA_CREDENTIALS]
        # END OF HACK
        # Get the username
        if self.KW_DN in credDict and credDict[self.KW_DN]:
            if self.KW_GROUP not in credDict:
                result = Registry.findDefaultGroupForDN(credDict[self.KW_DN])
                if not result["OK"]:
                    credDict[self.KW_USERNAME] = "anonymous"
                    credDict[self.KW_GROUP] = "visitor"
                else:
                    credDict[self.KW_GROUP] = result["Value"]
            if credDict[self.KW_GROUP] == self.KW_HOSTS_GROUP:
                # For host
                if not self.getHostNickName(credDict):
                    self.__authLogger.warn("Host is invalid")
                    if not allowAll:
                        return False
                    # If all, then set anon credentials
                    credDict[self.KW_USERNAME] = "anonymous"
                    credDict[self.KW_GROUP] = "visitor"
            else:
                # For users
                username = self.getUsername(credDict)
                suspended = self.isUserSuspended(credDict)
                if not username:
                    self.__authLogger.warn("User is invalid or does not belong to the group it's saying")
                if suspended:
                    self.__authLogger.warn("User is Suspended")

                if not username or suspended:
                    if not allowAll:
                        return False
                    # If all, then set anon credentials
                    credDict[self.KW_USERNAME] = "anonymous"
                    credDict[self.KW_GROUP] = "visitor"
        else:
            if not allowAll:
                return False
            credDict[self.KW_USERNAME] = "anonymous"
            credDict[self.KW_GROUP] = "visitor"

        # If any or all in the props, allow
        allowGroup = not validGroups or credDict[self.KW_GROUP] in validGroups
        if allowAll and allowGroup:
            return True
        # Check authorized groups
        if "authenticated" in lowerCaseProperties and allowGroup:
            return True
        if not self.matchProperties(credDict, requiredProperties):
            self.__authLogger.warn(
                "Client is not authorized\nValid properties: %s\nClient: %s" % (requiredProperties, credDict)
            )
            return False
        elif not allowGroup:
            self.__authLogger.warn("Client is not authorized\nValid groups: %s\nClient: %s" % (validGroups, credDict))
            return False
        return True

    def getHostNickName(self, credDict):
        """
        Discover the host nickname associated to the DN.
        The nickname will be included in the credentials dictionary.

        :type  credDict: dictionary
        :param credDict: Credentials to ckeck
        :return: Boolean specifying whether the nickname was found
        """
        if self.KW_DN not in credDict:
            return True
        if self.KW_GROUP not in credDict:
            return False
        retVal = Registry.getHostnameForDN(credDict[self.KW_DN])
        if not retVal["OK"]:
            gLogger.warn("Cannot find hostname for DN %s: %s" % (credDict[self.KW_DN], retVal["Message"]))
            return False
        credDict[self.KW_USERNAME] = retVal["Value"]
        credDict[self.KW_PROPERTIES] = Registry.getPropertiesForHost(credDict[self.KW_USERNAME], [])
        return True

    def getValidPropertiesForMethod(self, method, defaultProperties=False):
        """
        Get all authorized groups for calling a method

        :type  method: string
        :param method: Method to test
        :return: List containing the allowed groups
        """
        authProps = gConfig.getValue("%s/%s" % (self.authSection, method), [])
        if authProps:
            return authProps
        if defaultProperties:
            self.__authLogger.debug("Using hardcoded properties for method %s : %s" % (method, defaultProperties))
            if not isinstance(defaultProperties, (list, tuple)):
                return List.fromChar(defaultProperties)
            return defaultProperties
        defaultPath = "%s/Default" % "/".join(method.split("/")[:-1])
        authProps = gConfig.getValue("%s/%s" % (self.authSection, defaultPath), [])
        if authProps:
            self.__authLogger.debug("Method %s has no properties defined using %s" % (method, defaultPath))
            return authProps
        self.__authLogger.debug("Method %s has no authorization rules defined. Allowing no properties" % method)
        return []

    def getValidGroups(self, rawProperties):
        """Get valid groups as specified in the method authorization rules

        :param rawProperties: all method properties
        :type rawProperties: python:list
        :return: list of allowed groups or []
        """
        validGroups = []
        for prop in list(rawProperties):
            if prop.startswith("group:"):
                rawProperties.remove(prop)
                prop = prop.replace("group:", "")
                validGroups.append(prop)
            elif prop.startswith("vo:"):
                rawProperties.remove(prop)
                vo = prop.replace("vo:", "")
                result = Registry.getGroupsForVO(vo)
                if result["OK"]:
                    validGroups.extend(result["Value"])

        validGroups = list(set(validGroups))
        return validGroups

    def forwardedCredentials(self, credDict):
        """
        Check whether the credentials are being forwarded by a valid source

        :type  credDict: dictionary
        :param credDict: Credentials to ckeck
        :return: Boolean with the result
        """
        if self.KW_EXTRA_CREDENTIALS in credDict and isinstance(credDict[self.KW_EXTRA_CREDENTIALS], (tuple, list)):
            if self.KW_DN in credDict:
                retVal = Registry.getHostnameForDN(credDict[self.KW_DN])
                if retVal["OK"]:
                    hostname = retVal["Value"]
                    if Properties.TRUSTED_HOST in Registry.getPropertiesForHost(hostname, []):
                        return True
        return False

    def unpackForwardedCredentials(self, credDict):
        """
        Extract the forwarded credentials

        :type  credDict: dictionary
        :param credDict: Credentials to unpack
        """
        credDict[self.KW_DN] = credDict[self.KW_EXTRA_CREDENTIALS][0]
        credDict[self.KW_GROUP] = credDict[self.KW_EXTRA_CREDENTIALS][1]
        del credDict[self.KW_EXTRA_CREDENTIALS]

    def getUsername(self, credDict):
        """
        Discover the username associated to the DN. It will check if the selected group is valid.
        The username will be included in the credentials dictionary.

        :type  credDict: dictionary
        :param credDict: Credentials to check
        :return: Boolean specifying whether the username was found
        """
        if self.KW_DN not in credDict:
            return True
        if self.KW_GROUP not in credDict:
            result = Registry.findDefaultGroupForDN(credDict[self.KW_DN])
            if not result["OK"]:
                return False
            credDict[self.KW_GROUP] = result["Value"]
        credDict[self.KW_PROPERTIES] = Registry.getPropertiesForGroup(credDict[self.KW_GROUP], [])
        usersInGroup = Registry.getUsersInGroup(credDict[self.KW_GROUP], [])
        if not usersInGroup:
            return False
        retVal = Registry.getUsernameForDN(credDict[self.KW_DN], usersInGroup)
        if retVal["OK"]:
            credDict[self.KW_USERNAME] = retVal["Value"]
            return True
        return False

    def isUserSuspended(self, credDict):
        """Discover if the user is in Suspended status

        :param dict credDict: Credentials to check
        :return: Boolean True if user is Suspended
        """
        # Update credDict if the username is not there
        if self.KW_USERNAME not in credDict:
            self.getUsername(credDict)
        # If username or group is not known we can not judge if the user is suspended
        # These cases are treated elsewhere anyway
        if self.KW_USERNAME not in credDict or self.KW_GROUP not in credDict:
            return False
        suspendedVOList = Registry.getUserOption(credDict[self.KW_USERNAME], "Suspended", [])
        if not suspendedVOList:
            return False
        vo = Registry.getVOForGroup(credDict[self.KW_GROUP])
        if vo in suspendedVOList:
            return True
        return False

    def matchProperties(self, credDict, validProps, caseSensitive=False):
        """
        Return True if one or more properties are in the valid list of properties

        :type  props: list
        :param props: List of properties to match
        :type  validProps: list
        :param validProps: List of valid properties
        :return: Boolean specifying whether any property has matched the valid ones
        """

        # HACK: Map lower case properties to properties to make the check in lowercase but return the proper case
        if not caseSensitive:
            validProps = dict((prop.lower(), prop) for prop in validProps)
        else:
            validProps = dict((prop, prop) for prop in validProps)
        groupProperties = credDict[self.KW_PROPERTIES]
        foundProps = []
        for prop in groupProperties:
            if not caseSensitive:
                prop = prop.lower()
            if prop in validProps:
                foundProps.append(validProps[prop])
        credDict[self.KW_PROPERTIES] = foundProps
        return foundProps



"""
   DIRAC.ConfigurationSystem.Client.Helpers package
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.ConfigurationSystem.Client.Helpers.Path import *
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import getCSExtensions, getInstalledExtensions, getVO



""" Transformation Database Client Command Line Interface.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from DIRAC.Core.Base.Script import parseCommandLine

parseCommandLine()

from DIRAC.Core.Base.CLI import CLI
from DIRAC.Core.Base.API import API
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.TransformationSystem.Client.Transformation import Transformation
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

__RCSID__ = "$Id$"


def printDict(dictionary):
    """Dictionary pretty printing"""
    key_max = 0
    value_max = 0
    for key, value in dictionary.items():
        if len(key) > key_max:
            key_max = len(key)
        if len(str(value)) > value_max:
            value_max = len(str(value))
    for key, value in dictionary.items():
        print(key.rjust(key_max), " : ", str(value).ljust(value_max))


class TransformationCLI(CLI, API):
    def __init__(self):
        self.transClient = TransformationClient()
        self.indentSpace = 4
        CLI.__init__(self)
        API.__init__(self)

    def printPair(self, key, value, separator=":"):
        valueList = value.split("\n")
        print("%s%s%s %s" % (key, " " * (self.indentSpace - len(key)), separator, valueList[0].strip()))
        for valueLine in valueList[1:-1]:
            print("%s  %s" % (" " * self.indentSpace, valueLine.strip()))

    def do_help(self, args):
        """Default version of the help command
        Usage: help <command>
        OR use helpall to see description for all commands"""
        CLI.do_help(self, args)

    # overriting default help command
    def do_helpall(self, args):
        """
        Shows help information
            Usage: helpall <command>
            If no command is specified all commands are shown
        """
        if len(args) == 0:
            print("\nAvailable commands:\n")
            attrList = sorted(dir(self))
            for attribute in attrList:
                if attribute.find("do_") == 0:
                    self.printPair(attribute[3:], getattr(self, attribute).__doc__[1:])
                    print("")
        else:
            command = args.split()[0].strip()
            try:
                obj = getattr(self, "do_%s" % command)
            except Exception:
                print("There's no such %s command" % command)
                return
            self.printPair(command, obj.__doc__[1:])

    def do_shell(self, args):
        """Execute a shell command

        usage !<shell_command>
        """
        comm = args
        res = shellCall(0, comm)
        if res["OK"] and res["Value"][0] == 0:
            _returnCode, stdOut, stdErr = res["Value"]
            print("%s\n%s" % (stdOut, stdErr))
        else:
            print(res["Message"])

    def check_params(self, args, num):
        """Checks if the number of parameters correct"""
        argss = args.split()
        length = len(argss)
        if length < num:
            print("Error: Number of arguments provided %d less that required %d, please correct." % (length, num))
            return (False, length)
        return (argss, length)

    def check_id_or_name(self, id_or_name):
        """resolve name or Id by converting type of argument"""
        if id_or_name.isdigit():
            return int(id_or_name)  # its look like id
        return id_or_name

    ####################################################################
    #
    # These are the methods for transformation manipulation
    #

    def do_getall(self, args):
        """Get transformation details

        usage: getall [Status] [Status]
        """
        oTrans = Transformation()
        oTrans.getTransformations(transStatus=args.split(), printOutput=True)

    def do_getAllByUser(self, args):
        """Get all transformations created by a given user

        The first argument is the authorDN or username. The authorDN
        is preferred: it need to be inside quotes because contains
        white spaces. Only authorDN should be quoted.

        When the username is provided instead,
        the authorDN is retrieved from the uploaded proxy,
        so that the retrieved transformations are those created by
        the user who uploaded that proxy: that user could be different
        that the username provided to the function.

               usage: getAllByUser authorDN or username [Status] [Status]
        """
        oTrans = Transformation()
        argss = args.split()
        username = ""
        author = ""
        status = []
        if not len(argss) > 0:
            print(self.do_getAllByUser.__doc__)
            return

        # if the user didnt quoted the authorDN ends
        if "=" in argss[0] and argss[0][0] not in ["'", '"']:
            print("AuthorDN need to be quoted (just quote that argument)")
            return

        if argss[0][0] in ["'", '"']:  # authorDN given
            author = argss[0]
            status_idx = 1
            for arg in argss[1:]:
                author += " " + arg
                status_idx += 1
                if arg[-1] in ["'", '"']:
                    break
            # At this point we should have something like 'author'
            if not author[0] in ["'", '"'] or not author[-1] in ["'", '"']:
                print("AuthorDN need to be quoted (just quote that argument)")
                return
            else:
                author = author[1:-1]  # throw away the quotes
            # the rest are the requested status
            status = argss[status_idx:]
        else:  # username given
            username = argss[0]
            status = argss[1:]

        oTrans.getTransformationsByUser(authorDN=author, userName=username, transStatus=status, printOutput=True)

    def do_summaryTransformations(self, args):
        """Show the summary for a list of Transformations

        Fields starting with 'F' ('J')  refers to files (jobs).
        Proc. stand for processed.

            Usage: summaryTransformations <ProdID> [<ProdID> ...]
        """
        argss = args.split()
        if not len(argss) > 0:
            print(self.do_summaryTransformations.__doc__)
            return

        transid = argss
        oTrans = Transformation()
        oTrans.getSummaryTransformations(transID=transid)

    def do_getStatus(self, args):
        """Get transformation details

        usage: getStatus <transName|ID> [<transName2|ID> <transName3|ID> ...]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        for transName in argss:
            res = self.transClient.getTransformation(transName)
            if not res["OK"]:
                print("Getting status of %s failed: %s" % (transName, res["Message"]))
            else:
                print("%s: %s" % (transName, res["Value"]["Status"]))

    def do_setStatus(self, args):
        """Set transformation status

        usage: setStatus <Status> <transName|ID> [<transName2|ID> <transName3|ID> ...]
        Status <'New' 'Active' 'Stopped' 'Completed' 'Cleaning'>
        """
        argss = args.split()
        if not len(argss) > 1:
            print("transformation and status not supplied")
            return
        status = argss[0]
        transNames = argss[1:]
        for transName in transNames:
            res = self.transClient.setTransformationParameter(transName, "Status", status)
            if not res["OK"]:
                print("Setting status of %s failed: %s" % (transName, res["Message"]))
            else:
                print("%s set to %s" % (transName, status))

    def do_setGroupSize(self, args):
        """Set GroupSize of a transformation

        usage: setGroupSize <GroupSize> <transName|ID> [<transName2|ID> <transName3|ID> ...]
        """
        argss = args.split()
        if not len(argss) > 1:
            print("transformation and status not supplied")
            return
        groupSize = argss[0]
        transNames = argss[1:]
        for transName in transNames:
            res = self.transClient.setTransformationParameter(transName, "GroupSize", groupSize)
            if not res["OK"]:
                print("Setting GroupSize of %s failed: %s" % (transName, res["Message"]))
            else:
                print("%s set to %s" % (transName, groupSize))

    def do_start(self, args):
        """Start transformation

        usage: start <transName|ID> [<transName2|ID> <transName3|ID> ...]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        for transName in argss:
            res = self.transClient.setTransformationParameter(transName, "Status", "Active")
            if not res["OK"]:
                print("Setting Status of %s failed: %s" % (transName, res["Message"]))
            else:
                res = self.transClient.setTransformationParameter(transName, "AgentType", "Automatic")
                if not res["OK"]:
                    print("Setting AgentType of %s failed: %s" % (transName, res["Message"]))
                else:
                    print("%s started" % transName)

    def do_stop(self, args):
        """Stop transformation

        usage: stop <transID|ID> [<transName2|ID> <transName3|ID> ...]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        for transName in argss:
            res = self.transClient.setTransformationParameter(transName, "AgentType", "Manual")
            if not res["OK"]:
                print("Stopping of %s failed: %s" % (transName, res["Message"]))
            else:
                print("%s stopped" % transName)

    def do_flush(self, args):
        """Flush transformation

        usage: flush <transName|ID> [<transName2|ID> <transName3|ID> ...]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        for transName in argss:
            res = self.transClient.setTransformationParameter(transName, "Status", "Flush")
            if not res["OK"]:
                print("Flushing of %s failed: %s" % (transName, res["Message"]))
            else:
                print("%s flushing" % transName)

    def do_get(self, args):
        """Get transformation definition

        usage: get <transName|ID>
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        transName = argss[0]
        res = self.transClient.getTransformation(transName)
        if not res["OK"]:
            print("Failed to get %s: %s" % (transName, res["Message"]))
        else:
            res["Value"].pop("Body")
            printDict(res["Value"])

    def do_getBody(self, args):
        """Get transformation body

        usage: getBody <transName|ID>
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        transName = argss[0]
        res = self.transClient.getTransformation(transName)
        if not res["OK"]:
            print("Failed to get %s: %s" % (transName, res["Message"]))
        else:
            print(res["Value"]["Body"])

    def do_getFileStat(self, args):
        """Get transformation file statistics

        usage: getFileStat <transName|ID>
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        transName = argss[0]
        res = self.transClient.getTransformationStats(transName)
        if not res["OK"]:
            print("Failed to get statistics for %s: %s" % (transName, res["Message"]))
        else:
            res["Value"].pop("Total")
            printDict(res["Value"])

    def do_modMask(self, args):
        """Modify transformation input definition

        usage: modInput <mask> <transName|ID> [<transName2|ID> <transName3|ID> ...]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        mask = argss[0]
        transNames = argss[1:]
        for transName in transNames:
            res = self.transClient.setTransformationParameter(transName, "FileMask", mask)
            if not res["OK"]:
                print("Failed to modify input file mask for %s: %s" % (transName, res["Message"]))
            else:
                print("Updated %s filemask" % transName)

    def do_getFiles(self, args):
        """Get files for the transformation (optionally with a given status)

        usage: getFiles <transName|ID> [Status] [Status]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        transName = argss[0]
        status = argss[1:]
        res = self.transClient.getTransformation(transName)
        if not res["OK"]:
            print("Failed to get transformation information: %s" % res["Message"])
        else:
            selectDict = {"TransformationID": res["Value"]["TransformationID"]}
            if status:
                selectDict["Status"] = status
            res = self.transClient.getTransformationFiles(condDict=selectDict)
            if not res["OK"]:
                print("Failed to get transformation files: %s" % res["Message"])
            elif res["Value"]:
                self._printFormattedDictList(
                    res["Value"], ["LFN", "Status", "ErrorCount", "TargetSE", "LastUpdate"], "LFN", "LFN"
                )
            else:
                print("No files found")

    def do_getFileStatus(self, args):
        """Get file(s) status for the given transformation

        usage: getFileStatus <transName|ID> <lfn> [<lfn>...]
        """
        argss = args.split()
        if len(argss) < 2:
            print("transformation and file not supplied")
            return
        transName = argss[0]
        lfns = argss[1:]

        res = self.transClient.getTransformation(transName)
        if not res["OK"]:
            print("Failed to get transformation information: %s" % res["Message"])
        else:
            selectDict = {"TransformationID": res["Value"]["TransformationID"]}
            res = self.transClient.getTransformationFiles(condDict=selectDict)
            if not res["OK"]:
                print("Failed to get transformation files: %s" % res["Message"])
            elif res["Value"]:
                filesList = []
                for fileDict in res["Value"]:
                    if fileDict["LFN"] in lfns:
                        filesList.append(fileDict)
                if filesList:
                    self._printFormattedDictList(
                        filesList, ["LFN", "Status", "ErrorCount", "TargetSE", "LastUpdate"], "LFN", "LFN"
                    )
                else:
                    print("Could not find any LFN in", lfns, "for transformation", transName)
            else:
                print("No files found")

    def do_getOutputFiles(self, args):
        """Get output files for the transformation

        usage: getOutputFiles <transName|ID>
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        transName = argss[0]
        res = self.transClient.getTransformation(transName)
        if not res["OK"]:
            print("Failed to get transformation information: %s" % res["Message"])
        else:
            fc = FileCatalog()
            meta = {}
            meta["ProdID"] = transName
            res = fc.findFilesByMetadata(meta)
            if not res["OK"]:
                print(res["Message"])
                return
            if not len(res["Value"]) > 0:
                print("No output files yet for transformation %d" % int(transName))
                return
            else:
                for lfn in res["Value"]:
                    print(lfn)

    def do_getInputDataQuery(self, args):
        """Get input data query for the transformation

        usage: getInputDataQuery <transName|ID>
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no transformation supplied")
            return
        transName = argss[0]
        # res = self.transClient.getTransformationInputDataQuery( transName )
        res = self.transClient.getTransformationMetaQuery(transName, "Input")
        if not res["OK"]:
            print("Failed to get transformation input data query: %s" % res["Message"])
        else:
            print(res["Value"])

    def do_setFileStatus(self, args):
        """Set file status for the given transformation

        usage: setFileStatus <transName|ID> <lfn> <status>
        """
        argss = args.split()
        if not len(argss) == 3:
            print("transformation file and status not supplied")
            return
        transName = argss[0]
        lfn = argss[1]
        status = argss[2]
        res = self.transClient.setFileStatusForTransformation(transName, status, [lfn])
        if not res["OK"]:
            print("Failed to update file status: %s" % res["Message"])
        else:
            print("Updated file status to %s" % status)

    def do_resetFile(self, args):
        """Reset file status for the given transformation

        usage: resetFile <transName|ID> <lfns>
        """
        argss = args.split()
        if not len(argss) > 1:
            print("transformation and file(s) not supplied")
            return
        transName = argss[0]
        lfns = argss[1:]
        res = self.transClient.setFileStatusForTransformation(transName, TransformationFilesStatus.UNUSED, lfns)
        if not res["OK"]:
            print("Failed to reset file status: %s" % res["Message"])
        else:
            if "Failed" in res["Value"]:
                print("Could not reset some files: ")
                for lfn, reason in res["Value"]["Failed"].items():
                    print(lfn, reason)
            else:
                print("Updated file statuses to 'Unused' for %d file(s)" % len(lfns))

    def do_resetProcessedFile(self, args):
        """Reset file status for the given transformation
        usage: resetFile <transName|ID> <lfn>
        """
        argss = args.split()

        if not len(argss) > 1:
            print("transformation and file(s) not supplied")
            return
        transName = argss[0]
        lfns = argss[1:]
        res = self.transClient.setFileStatusForTransformation(
            transName, TransformationFilesStatus.UNUSED, lfns, force=True
        )
        if not res["OK"]:
            print("Failed to reset file status: %s" % res["Message"])
        else:
            if "Failed" in res["Value"] and res["Value"]["Failed"]:
                print("Could not reset some files: ")
                for lfn, reason in res["Value"]["Failed"].items():
                    print(lfn, reason)
            else:
                print("Updated file statuses to 'Unused' for %d file(s)" % len(lfns))

    ####################################################################
    #
    # These are the methods for file manipulation
    #

    def do_addDirectory(self, args):
        """Add files from the given catalog directory

        usage: addDirectory <directory> [directory]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no directory supplied")
            return
        for directory in argss:
            res = self.transClient.addDirectory(directory, force=True)
            if not res["OK"]:
                print("failed to add directory %s: %s" % (directory, res["Message"]))
            else:
                print("added %s files for %s" % (res["Value"], directory))

    def do_replicas(self, args):
        """Get replicas for <path>

        usage: replicas <lfn> [lfn]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no files supplied")
            return
        res = self.transClient.getReplicas(argss)
        if not res["OK"]:
            print("failed to get any replica information: %s" % res["Message"])
            return
        for lfn in sorted(res["Value"]["Failed"]):
            error = res["Value"]["Failed"][lfn]
            print("failed to get replica information for %s: %s" % (lfn, error))
        for lfn in sorted(res["Value"]["Successful"]):
            ses = sorted(res["Value"]["Successful"][lfn])
            outStr = "%s :" % lfn.ljust(100)
            for se in ses:
                outStr = "%s %s" % (outStr, se.ljust(15))
            print(outStr)

    def do_addFile(self, args):
        """Add new files to transformation DB

        usage: addFile <lfn> [lfn]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no files supplied")
            return
        lfnDict = {}
        for lfn in argss:
            lfnDict[lfn] = {
                "PFN": "IGNORED-PFN",
                "SE": "IGNORED-SE",
                "Size": 0,
                "GUID": "IGNORED-GUID",
                "Checksum": "IGNORED-CHECKSUM",
            }
        res = self.transClient.addFile(lfnDict, force=True)
        if not res["OK"]:
            print("failed to add any files: %s" % res["Message"])
            return
        for lfn in sorted(res["Value"]["Failed"]):
            error = res["Value"]["Failed"][lfn]
            print("failed to add %s: %s" % (lfn, error))
        for lfn in sorted(res["Value"]["Successful"]):
            print("added %s" % lfn)

    def do_removeFile(self, args):
        """Remove file from transformation DB

        usage: removeFile <lfn> [lfn]
        """
        argss = args.split()
        if not len(argss) > 0:
            print("no files supplied")
            return
        res = self.transClient.removeFile(argss)
        if not res["OK"]:
            print("failed to remove any files: %s" % res["Message"])
            return
        for lfn in sorted(res["Value"]["Failed"]):
            error = res["Value"]["Failed"][lfn]
            print("failed to remove %s: %s" % (lfn, error))
        for lfn in sorted(res["Value"]["Successful"]):
            print("removed %s" % lfn)

    def do_addReplica(self, args):
        """Add new replica to the transformation DB

        usage: addReplica <lfn> <se>
        """
        argss = args.split()
        if not len(argss) == 2:
            print("no file info supplied")
            return
        lfn = argss[0]
        se = argss[1]
        lfnDict = {}
        lfnDict[lfn] = {
            "PFN": "IGNORED-PFN",
            "SE": se,
            "Size": 0,
            "GUID": "IGNORED-GUID",
            "Checksum": "IGNORED-CHECKSUM",
        }
        res = self.transClient.addReplica(lfnDict, force=True)
        if not res["OK"]:
            print("failed to add replica: %s" % res["Message"])
            return
        for lfn in sorted(res["Value"]["Failed"]):
            error = res["Value"]["Failed"][lfn]
            print("failed to add replica: %s" % (error))
        for lfn in sorted(res["Value"]["Successful"]):
            print("added %s" % lfn)

    def do_removeReplica(self, args):
        """Remove replica from the transformation DB

        usage: removeReplica <lfn> <se>
        """
        argss = args.split()
        if not len(argss) == 2:
            print("no file info supplied")
            return
        lfn = argss[0]
        se = argss[1]
        lfnDict = {}
        lfnDict[lfn] = {
            "PFN": "IGNORED-PFN",
            "SE": se,
            "Size": 0,
            "GUID": "IGNORED-GUID",
            "Checksum": "IGNORED-CHECKSUM",
        }
        res = self.transClient.removeReplica(lfnDict)
        if not res["OK"]:
            print("failed to remove replica: %s" % res["Message"])
            return
        for lfn in sorted(res["Value"]["Failed"]):
            error = res["Value"]["Failed"][lfn]
            print("failed to remove replica: %s" % (error))
        for lfn in sorted(res["Value"]["Successful"]):
            print("removed %s" % lfn)

    def do_setReplicaStatus(self, args):
        """Set replica status, usually used to mark a replica Problematic

        usage: setReplicaStatus <lfn> <status> <se>
        """
        argss = args.split()
        if not len(argss) > 2:
            print("no file info supplied")
            return
        lfn = argss[0]
        status = argss[1]
        se = argss[2]
        lfnDict = {}
        lfnDict[lfn] = {
            "Status": status,
            "PFN": "IGNORED-PFN",
            "SE": se,
            "Size": 0,
            "GUID": "IGNORED-GUID",
            "Checksum": "IGNORED-CHECKSUM",
        }
        res = self.transClient.setReplicaStatus(lfnDict)
        if not res["OK"]:
            print("failed to set replica status: %s" % res["Message"])
            return
        for lfn in sorted(res["Value"]["Failed"]):
            error = res["Value"]["Failed"][lfn]
            print("failed to set replica status: %s" % (error))
        for lfn in sorted(res["Value"]["Successful"]):
            print("updated replica status %s" % lfn)


if __name__ == "__main__":
    cli = TransformationCLI()
    cli.cmdloop()



""" SSH (Virtual) Computing Element

For a given IP/host it will send jobs directly through ssh

**Configuration Parameters**

Configuration for the SSHComputingElement submission can be done via the configuration system.

BatchSystem:
   Underlying batch system that is going to be used to orchestrate executable files. The Batch System has to be
   accessible from the LocalCE. By default, the LocalComputingElement submits directly on the host via the Host class.

SharedArea:
   Area used to store executable/output/error files if they are not aready defined via BatchOutput, BatchError,
   InfoArea, ExecutableArea and/or WorkArea. The path should be absolute.

BatchOutput:
   Area where the job outputs are stored.
   If not defined: SharedArea + '/data' is used.
   If not absolute: SharedArea + path is used.

BatchError:
   Area where the job errors are stored.
   If not defined: SharedArea + '/data' is used.
   If not absolute: SharedArea + path is used.

ExecutableArea:
   Area where the executable files are stored if necessary.
   If not defined: SharedArea + '/data' is used.
   If not absolute: SharedArea + path is used.

SSHHost:
   SSH host name

SSHUser:
   SSH user login

SSHPassword:
   SSH password

SSHPort:
   Port number if not standard, e.g. for the gsissh access

SSHKey:
   Location of the ssh private key for no-password connection

SSHOptions:
   Any other SSH options to be used

SSHTunnel:
   String defining the use of intermediate SSH host. Example::

     ssh -i /private/key/location -l final_user final_host

SSHType:
   SSH (default) or gsissh

**Code Documentation**

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import json
import stat
import shutil
import errno
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import quote as urlquote
from six.moves.urllib.parse import unquote as urlunquote
from six.moves import shlex_quote

import DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC import rootPath
from DIRAC import gLogger

from DIRAC.Resources.Computing.ComputingElement import ComputingElement
from DIRAC.Resources.Computing.PilotBundle import bundleProxy, writeScript
from DIRAC.Resources.Computing.BatchSystems.executeBatch import executeBatchContent
from DIRAC.Core.Utilities.List import uniqueElements
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.Core.Utilities.List import breakListIntoChunks


class SSH(object):
    """SSH class encapsulates passing commands and files through an SSH tunnel
    to a remote host. It can use either ssh or gsissh access. The final host
    where the commands will be executed and where the files will copied/retrieved
    can be reached through an intermediate host if SSHTunnel parameters is defined.

    SSH constructor parameters are defined in a SSH accessible Computing Element
    in the Configuration System:

    - SSHHost: SSH host name
    - SSHUser: SSH user login
    - SSHPassword: SSH password
    - SSHPort: port number if not standard, e.g. for the gsissh access
    - SSHKey: location of the ssh private key for no-password connection
    - SSHOptions: any other SSH options to be used
    - SSHTunnel: string defining the use of intermediate SSH host. Example:
                 'ssh -i /private/key/location -l final_user final_host'
    - SSHType: ssh ( default ) or gsissh

    The class public interface includes two methods:

    sshCall( timeout, command_sequence )
    scpCall( timeout, local_file, remote_file, upload = False/True )
    """

    def __init__(self, host=None, parameters=None):

        self.host = host
        if parameters is None:
            parameters = {}
        if not host:
            self.host = parameters.get("SSHHost", "")

        self.user = parameters.get("SSHUser", "")
        self.password = parameters.get("SSHPassword", "")
        self.port = parameters.get("SSHPort", "")
        self.key = parameters.get("SSHKey", "")
        self.options = parameters.get("SSHOptions", "")
        self.sshTunnel = parameters.get("SSHTunnel", "")
        self.sshType = parameters.get("SSHType", "ssh")

        if self.port:
            self.options += " -p %s" % self.port
        if self.key:
            self.options += " -i %s" % self.key
        self.options = self.options.strip()

        self.log = gLogger.getSubLogger("SSH")

    def __ssh_call(self, command, timeout):

        try:
            import pexpect

            expectFlag = True
        except BaseException:
            from DIRAC.Core.Utilities.Subprocess import shellCall

            expectFlag = False

        if not timeout:
            timeout = 999

        if expectFlag:
            ssh_newkey = "Are you sure you want to continue connecting"
            try:
                child = pexpect.spawn(command, timeout=timeout, encoding="utf-8")
                i = child.expect([pexpect.TIMEOUT, ssh_newkey, pexpect.EOF, "assword: "])
                if i == 0:  # Timeout
                    return S_OK((-1, child.before, "SSH login failed"))
                elif i == 1:  # SSH does not have the public key. Just accept it.
                    child.sendline("yes")
                    child.expect("assword: ")
                    i = child.expect([pexpect.TIMEOUT, "assword: "])
                    if i == 0:  # Timeout
                        return S_OK((-1, str(child.before) + str(child.after), "SSH login failed"))
                    elif i == 1:
                        child.sendline(self.password)
                        child.expect(pexpect.EOF)
                        return S_OK((0, child.before, ""))
                elif i == 2:
                    # Passwordless login, get the output
                    return S_OK((0, child.before, ""))

                if self.password:
                    child.sendline(self.password)
                    child.expect(pexpect.EOF)
                    return S_OK((0, child.before, ""))
                return S_ERROR((-2, child.before, ""))
            except Exception as x:
                res = (-1, "Encountered exception %s: %s" % (Exception, str(x)))
                return S_ERROR(res)
        else:
            # Try passwordless login
            result = shellCall(timeout, command)
            #      print ( "!!! SSH command: %s returned %s\n" % (command, result) )
            if result["Value"][0] == 255:
                return S_ERROR((-1, "Cannot connect to host %s" % self.host, ""))
            return result

    def sshCall(self, timeout, cmdSeq):
        """Execute remote command via a ssh remote call

        :param int timeout: timeout of the command
        :param cmdSeq: list of command components
        :type cmdSeq: python:list
        """

        command = cmdSeq
        if isinstance(cmdSeq, list):
            command = " ".join(cmdSeq)

        pattern = "__DIRAC__"

        if self.sshTunnel:
            command = command.replace("'", '\\\\\\"')
            command = command.replace("$", "\\\\\\$")
            command = '/bin/sh -c \' %s -q %s -l %s %s "%s \\"echo %s; %s\\" " \' ' % (
                self.sshType,
                self.options,
                self.user,
                self.host,
                self.sshTunnel,
                pattern,
                command,
            )
        else:
            # command = command.replace( '$', '\$' )
            command = '%s -q %s -l %s %s "echo %s; %s"' % (
                self.sshType,
                self.options,
                self.user,
                self.host,
                pattern,
                command,
            )
        self.log.debug("SSH command: %s" % command)
        result = self.__ssh_call(command, timeout)
        self.log.debug("SSH command result %s" % str(result))
        if not result["OK"]:
            return result

        # Take the output only after the predefined pattern
        ind = result["Value"][1].find("__DIRAC__")
        if ind == -1:
            return result

        status, output, error = result["Value"]
        output = output[ind + 9 :]
        if output.startswith("\r"):
            output = output[1:]
        if output.startswith("\n"):
            output = output[1:]

        result["Value"] = (status, output, error)
        return result

    def scpCall(self, timeout, localFile, remoteFile, postUploadCommand="", upload=True):
        """Perform file copy through an SSH magic.

        :param int timeout: timeout of the command
        :param str localFile: local file path, serves as source for uploading and destination for downloading.
                              Can take 'Memory' as value, in this case the downloaded contents is returned
                              as result['Value']
        :param str remoteFile: remote file full path
        :param str postUploadCommand: command executed on the remote side after file upload
        :param bool upload: upload if True, download otherwise
        """
        # shlex_quote aims to prevent any security issue or problems with filepath containing spaces
        # it returns a shell-escaped version of the filename
        localFile = shlex_quote(localFile)
        remoteFile = shlex_quote(remoteFile)
        if upload:
            if self.sshTunnel:
                remoteFile = remoteFile.replace("$", r"\\\\\$")
                postUploadCommand = postUploadCommand.replace("$", r"\\\\\$")
                command = '/bin/sh -c \'cat %s | %s -q %s %s@%s "%s \\"cat > %s; %s\\""\' ' % (
                    localFile,
                    self.sshType,
                    self.options,
                    self.user,
                    self.host,
                    self.sshTunnel,
                    remoteFile,
                    postUploadCommand,
                )
            else:
                command = "/bin/sh -c \"cat %s | %s -q %s %s@%s 'cat > %s; %s'\" " % (
                    localFile,
                    self.sshType,
                    self.options,
                    self.user,
                    self.host,
                    remoteFile,
                    postUploadCommand,
                )
        else:
            finalCat = "| cat > %s" % localFile
            if localFile.lower() == "memory":
                finalCat = ""
            if self.sshTunnel:
                remoteFile = remoteFile.replace("$", "\\\\\\$")
                command = '/bin/sh -c \'%s -q %s -l %s %s "%s \\"cat %s\\"" %s\'' % (
                    self.sshType,
                    self.options,
                    self.user,
                    self.host,
                    self.sshTunnel,
                    remoteFile,
                    finalCat,
                )
            else:
                remoteFile = remoteFile.replace("$", r"\$")
                command = "/bin/sh -c '%s -q %s -l %s %s \"cat %s\" %s'" % (
                    self.sshType,
                    self.options,
                    self.user,
                    self.host,
                    remoteFile,
                    finalCat,
                )

        self.log.debug("SSH copy command: %s" % command)
        return self.__ssh_call(command, timeout)


class SSHComputingElement(ComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(SSHComputingElement, self).__init__(ceUniqueID)

        self.ceType = "SSH"
        self.execution = "SSHCE"
        self.submittedJobs = 0
        self.outputTemplate = ""
        self.errorTemplate = ""

    ############################################################################
    def setProxy(self, proxy, valid=0):
        """
        Set and prepare proxy to use

        :param str proxy: proxy to use
        :param int valid: proxy validity period
        :return:  S_OK/S_ERROR
        """
        ComputingElement.setProxy(self, proxy, valid)
        if self.ceParameters.get("SSHType", "ssh") == "gsissh":
            result = self._prepareProxy()
            if not result["OK"]:
                gLogger.error("SSHComputingElement: failed to set up proxy", result["Message"])
                return result
        return S_OK()

    #############################################################################
    def _addCEConfigDefaults(self):
        """Method to make sure all necessary Configuration Parameters are defined"""
        # First assure that any global parameters are loaded
        ComputingElement._addCEConfigDefaults(self)
        # Now batch system specific ones
        if "ExecQueue" not in self.ceParameters:
            self.ceParameters["ExecQueue"] = self.ceParameters.get("Queue", "")

        if "SharedArea" not in self.ceParameters:
            # . isn't a good location, move to $HOME
            self.ceParameters["SharedArea"] = "$HOME"

        if "BatchOutput" not in self.ceParameters:
            self.ceParameters["BatchOutput"] = "data"

        if "BatchError" not in self.ceParameters:
            self.ceParameters["BatchError"] = "data"

        if "ExecutableArea" not in self.ceParameters:
            self.ceParameters["ExecutableArea"] = "data"

        if "InfoArea" not in self.ceParameters:
            self.ceParameters["InfoArea"] = "info"

        if "WorkArea" not in self.ceParameters:
            self.ceParameters["WorkArea"] = "work"

    def _reset(self):
        """Process CE parameters and make necessary adjustments"""
        batchSystemName = self.ceParameters.get("BatchSystem", "Host")
        if "BatchSystem" not in self.ceParameters:
            self.ceParameters["BatchSystem"] = batchSystemName
        result = self.loadBatchSystem(batchSystemName)
        if not result["OK"]:
            self.log.error("Failed to load the batch system plugin", batchSystemName)
            return result

        self.user = self.ceParameters["SSHUser"]
        self.queue = self.ceParameters["Queue"]
        self.submitOptions = self.ceParameters.get("SubmitOptions", "")
        if "ExecQueue" not in self.ceParameters or not self.ceParameters["ExecQueue"]:
            self.ceParameters["ExecQueue"] = self.ceParameters.get("Queue", "")
        self.execQueue = self.ceParameters["ExecQueue"]
        self.log.info("Using queue: ", self.queue)

        self.sharedArea = self.ceParameters["SharedArea"]
        self.batchOutput = self.ceParameters["BatchOutput"]
        if not self.batchOutput.startswith("/"):
            self.batchOutput = os.path.join(self.sharedArea, self.batchOutput)
        self.batchError = self.ceParameters["BatchError"]
        if not self.batchError.startswith("/"):
            self.batchError = os.path.join(self.sharedArea, self.batchError)
        self.infoArea = self.ceParameters["InfoArea"]
        if not self.infoArea.startswith("/"):
            self.infoArea = os.path.join(self.sharedArea, self.infoArea)
        self.executableArea = self.ceParameters["ExecutableArea"]
        if not self.executableArea.startswith("/"):
            self.executableArea = os.path.join(self.sharedArea, self.executableArea)
        self.workArea = self.ceParameters["WorkArea"]
        if not self.workArea.startswith("/"):
            self.workArea = os.path.join(self.sharedArea, self.workArea)

        self.submitOptions = self.ceParameters.get("SubmitOptions", "")
        self.removeOutput = True
        if "RemoveOutput" in self.ceParameters:
            if self.ceParameters["RemoveOutput"].lower() in ["no", "false", "0"]:
                self.removeOutput = False
        self.preamble = self.ceParameters.get("Preamble", "")

        result = self._prepareRemoteHost()
        if not result["OK"]:
            return result

        return S_OK()

    def _prepareRemoteHost(self, host=None):
        """Prepare remote directories and upload control script"""

        ssh = SSH(host=host, parameters=self.ceParameters)

        # Make remote directories
        dirTuple = tuple(
            uniqueElements(
                [self.sharedArea, self.executableArea, self.infoArea, self.batchOutput, self.batchError, self.workArea]
            )
        )
        nDirs = len(dirTuple)
        cmd = "mkdir -p %s; " * nDirs % dirTuple
        cmd = "bash -c '%s'" % cmd
        self.log.verbose("Creating working directories on %s" % self.ceParameters["SSHHost"])
        result = ssh.sshCall(30, cmd)
        if not result["OK"]:
            self.log.error("Failed creating working directories", "(%s)" % result["Message"][1])
            return result
        status, output, _error = result["Value"]
        if status == -1:
            self.log.error("Timeout while creating directories")
            return S_ERROR(errno.ETIME, "Timeout while creating directories")
        if "cannot" in output:
            self.log.error("Failed to create directories", "(%s)" % output)
            return S_ERROR(errno.EACCES, "Failed to create directories")

        # Upload the control script now
        result = self._generateControlScript()
        if not result["OK"]:
            self.log.warn("Failed generating control script")
            return result
        localScript = result["Value"]
        self.log.verbose(
            "Uploading %s script to %s" % (self.batchSystem.__class__.__name__, self.ceParameters["SSHHost"])
        )
        remoteScript = "%s/execute_batch" % self.sharedArea
        result = ssh.scpCall(30, localScript, remoteScript, postUploadCommand="chmod +x %s" % remoteScript)
        if not result["OK"]:
            self.log.warn("Failed uploading control script: %s" % result["Message"][1])
            return result
        status, output, _error = result["Value"]
        if status != 0:
            if status == -1:
                self.log.warn("Timeout while uploading control script")
                return S_ERROR("Timeout while uploading control script")
            self.log.warn("Failed uploading control script: %s" % output)
            return S_ERROR("Failed uploading control script")

        # Delete the generated control script locally
        try:
            os.remove(localScript)
        except OSError:
            self.log.warn("Failed removing the generated control script locally")
            return S_ERROR("Failed removing the generated control script locally")

        # Chmod the control scripts
        # self.log.verbose( 'Chmod +x control script' )
        # result = ssh.sshCall( 10, "chmod +x %s/%s" % ( self.sharedArea, self.controlScript ) )
        # if not result['OK']:
        #  self.log.warn( 'Failed chmod control script: %s' % result['Message'][1] )
        #  return result
        # status, output, _error = result['Value']
        # if status != 0:
        #  if status == -1:
        #    self.log.warn( 'Timeout while chmod control script' )
        #    return S_ERROR( 'Timeout while chmod control script' )
        #  else:
        #    self.log.warn( 'Failed uploading chmod script: %s' % output )
        #    return S_ERROR( 'Failed uploading chmod script' )

        return S_OK()

    def _generateControlScript(self):
        """Generates a control script from a BatchSystem and a script called executeBatch

        :return: a path containing the script generated
        """
        # Get the batch system module to use
        batchSystemDir = os.path.join(os.path.dirname(DIRAC.__file__), "Resources", "Computing", "BatchSystems")
        batchSystemScript = os.path.join(batchSystemDir, "%s.py" % self.batchSystem.__class__.__name__)

        # Get the executeBatch.py content: an str variable composed of code content that has to be extracted
        # The control script is generated from the batch system module and this variable
        controlScript = os.path.join(batchSystemDir, "control_script.py")

        try:
            shutil.copyfile(batchSystemScript, controlScript)
            with open(controlScript, "a") as cs:
                cs.write(executeBatchContent)
        except IOError:
            return S_ERROR("IO Error trying to generate control script")

        return S_OK("%s" % controlScript)

    def __executeHostCommand(self, command, options, ssh=None, host=None):

        if not ssh:
            ssh = SSH(host=host, parameters=self.ceParameters)

        options["BatchSystem"] = self.batchSystem.__class__.__name__
        options["Method"] = command
        options["SharedDir"] = self.sharedArea
        options["OutputDir"] = self.batchOutput
        options["ErrorDir"] = self.batchError
        options["WorkDir"] = self.workArea
        options["InfoDir"] = self.infoArea
        options["ExecutionContext"] = self.execution
        options["User"] = self.user
        options["Queue"] = self.queue

        options = json.dumps(options)
        options = urlquote(options)

        cmd = (
            "bash --login -c 'python %s/execute_batch %s || python3 %s/execute_batch %s || python2 %s/execute_batch %s'"
            % (self.sharedArea, options, self.sharedArea, options, self.sharedArea, options)
        )

        self.log.verbose("CE submission command: %s" % cmd)

        result = ssh.sshCall(120, cmd)
        if not result["OK"]:
            self.log.error("%s CE job submission failed" % self.ceType, result["Message"])
            return result

        sshStatus = result["Value"][0]
        sshStdout = result["Value"][1]
        sshStderr = result["Value"][2]

        # Examine results of the job submission
        if sshStatus == 0:
            output = sshStdout.strip().replace("\r", "").strip()
            try:
                index = output.index("============= Start output ===============")
                output = output[index + 42 :]
            except Exception:
                return S_ERROR("Invalid output from remote command: %s" % output)
            try:
                output = urlunquote(output)
                result = json.loads(output)
                if isinstance(result, six.string_types) and result.startswith("Exception:"):
                    return S_ERROR(result)
                return S_OK(result)
            except Exception:
                return S_ERROR("Invalid return structure from job submission")
        else:
            return S_ERROR("\n".join([sshStdout, sshStderr]))

    def submitJob(self, executableFile, proxy, numberOfJobs=1):

        #    self.log.verbose( "Executable file path: %s" % executableFile )
        if not os.access(executableFile, 5):
            os.chmod(executableFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

        # if no proxy is supplied, the executable can be submitted directly
        # otherwise a wrapper script is needed to get the proxy to the execution node
        # The wrapper script makes debugging more complicated and thus it is
        # recommended to transfer a proxy inside the executable if possible.
        if proxy:
            self.log.verbose("Setting up proxy for payload")
            wrapperContent = bundleProxy(executableFile, proxy)
            name = writeScript(wrapperContent, os.getcwd())
            submitFile = name
        else:  # no proxy
            submitFile = executableFile

        result = self._submitJobToHost(submitFile, numberOfJobs)
        if proxy:
            os.remove(submitFile)

        return result

    def _submitJobToHost(self, executableFile, numberOfJobs, host=None):
        """Submit prepared executable to the given host"""
        ssh = SSH(host=host, parameters=self.ceParameters)
        # Copy the executable
        submitFile = os.path.join(self.executableArea, os.path.basename(executableFile))
        result = ssh.scpCall(30, executableFile, submitFile, postUploadCommand="chmod +x %s" % submitFile)
        if not result["OK"]:
            return result

        jobStamps = []
        for _i in range(numberOfJobs):
            jobStamps.append(makeGuid()[:8])

        numberOfProcessors = self.ceParameters.get("NumberOfProcessors", 1)
        wholeNode = self.ceParameters.get("WholeNode", False)
        # numberOfNodes is treated as a string as it can contain values such as "2-4"
        # where 2 would represent the minimum number of nodes to allocate, and 4 the maximum
        numberOfNodes = self.ceParameters.get("NumberOfNodes", "1")
        self.numberOfGPUs = self.ceParameters.get("NumberOfGPUs")

        # Collect command options
        commandOptions = {
            "Executable": submitFile,
            "NJobs": numberOfJobs,
            "SubmitOptions": self.submitOptions,
            "JobStamps": jobStamps,
            "WholeNode": wholeNode,
            "NumberOfProcessors": numberOfProcessors,
            "NumberOfNodes": numberOfNodes,
            "Preamble": self.preamble,
            "NumberOfGPUs": self.numberOfGPUs,
        }
        if host:
            commandOptions["SSHNodeHost"] = host

        resultCommand = self.__executeHostCommand("submitJob", commandOptions, ssh=ssh, host=host)
        if not resultCommand["OK"]:
            return resultCommand

        result = resultCommand["Value"]
        if result["Status"] != 0:
            return S_ERROR("Failed job submission: %s" % result["Message"])
        else:
            batchIDs = result["Jobs"]
            if batchIDs:
                batchSystemName = self.batchSystem.__class__.__name__.lower()
                if host is None:
                    jobIDs = [
                        "%s%s://%s/%s" % (self.ceType.lower(), batchSystemName, self.ceName, _id) for _id in batchIDs
                    ]
                else:
                    jobIDs = [
                        "%s%s://%s/%s/%s" % (self.ceType.lower(), batchSystemName, self.ceName, host, _id)
                        for _id in batchIDs
                    ]
            else:
                return S_ERROR("No jobs IDs returned")

        result = S_OK(jobIDs)
        self.submittedJobs += len(batchIDs)

        return result

    def killJob(self, jobIDList):
        """Kill a bunch of jobs"""
        if isinstance(jobIDList, six.string_types):
            jobIDList = [jobIDList]
        return self._killJobOnHost(jobIDList)

    def _killJobOnHost(self, jobIDList, host=None):
        """Kill the jobs for the given list of job IDs"""
        jobDict = {}
        for job in jobIDList:
            stamp = os.path.basename(urlparse(job).path)
            jobDict[stamp] = job
        stampList = list(jobDict)

        commandOptions = {"JobIDList": stampList, "User": self.user}
        resultCommand = self.__executeHostCommand("killJob", commandOptions, host=host)
        if not resultCommand["OK"]:
            return resultCommand

        result = resultCommand["Value"]
        if result["Status"] != 0:
            return S_ERROR("Failed job kill: %s" % result["Message"])

        if result["Failed"]:
            return S_ERROR("%d jobs failed killing" % len(result["Failed"]))

        return S_OK(len(result["Successful"]))

    def _getHostStatus(self, host=None):
        """Get jobs running at a given host"""
        resultCommand = self.__executeHostCommand("getCEStatus", {}, host=host)
        if not resultCommand["OK"]:
            return resultCommand

        result = resultCommand["Value"]
        if result["Status"] != 0:
            return S_ERROR("Failed to get CE status: %s" % result["Message"])

        return S_OK(result)

    def getCEStatus(self):
        """Method to return information on running and pending jobs."""
        result = S_OK()
        result["SubmittedJobs"] = self.submittedJobs
        result["RunningJobs"] = 0
        result["WaitingJobs"] = 0

        resultHost = self._getHostStatus()
        if not resultHost["OK"]:
            return resultHost

        result["RunningJobs"] = resultHost["Value"].get("Running", 0)
        result["WaitingJobs"] = resultHost["Value"].get("Waiting", 0)
        if "AvailableCores" in resultHost["Value"]:
            result["AvailableCores"] = resultHost["Value"]["AvailableCores"]
        self.log.verbose("Waiting Jobs: ", result["WaitingJobs"])
        self.log.verbose("Running Jobs: ", result["RunningJobs"])

        return result

    def getJobStatus(self, jobIDList):
        """Get the status information for the given list of jobs"""
        return self._getJobStatusOnHost(jobIDList)

    def _getJobStatusOnHost(self, jobIDList, host=None):
        """Get the status information for the given list of jobs"""

        resultDict = {}
        jobDict = {}
        for job in jobIDList:
            stamp = os.path.basename(urlparse(job).path)
            jobDict[stamp] = job
        stampList = list(jobDict)

        for jobList in breakListIntoChunks(stampList, 100):
            resultCommand = self.__executeHostCommand("getJobStatus", {"JobIDList": jobList}, host=host)
            if not resultCommand["OK"]:
                return resultCommand

            result = resultCommand["Value"]
            if result["Status"] != 0:
                return S_ERROR("Failed to get job status: %s" % result["Message"])

            for stamp in result["Jobs"]:
                resultDict[jobDict[stamp]] = result["Jobs"][stamp]

        return S_OK(resultDict)

    def _getJobOutputFiles(self, jobID, host=None):
        """Get output file names for the specific CE"""
        jobStamp = os.path.basename(urlparse(jobID).path)
        host = urlparse(jobID).hostname

        if "OutputTemplate" in self.ceParameters:
            self.outputTemplate = self.ceParameters["OutputTemplate"]
            self.errorTemplate = self.ceParameters["ErrorTemplate"]

        if self.outputTemplate:
            output = self.outputTemplate % jobStamp
            error = self.errorTemplate % jobStamp
        elif "OutputTemplate" in self.ceParameters:
            self.outputTemplate = self.ceParameters["OutputTemplate"]
            self.errorTemplate = self.ceParameters["ErrorTemplate"]
            output = self.outputTemplate % jobStamp
            error = self.errorTemplate % jobStamp
        elif hasattr(self.batchSystem, "getJobOutputFiles"):
            # numberOfNodes is treated as a string as it can contain values such as "2-4"
            # where 2 would represent the minimum number of nodes to allocate, and 4 the maximum
            numberOfNodes = self.ceParameters.get("NumberOfNodes", "1")
            commandOptions = {
                "JobIDList": [jobStamp],
                "OutputDir": self.batchOutput,
                "ErrorDir": self.batchError,
                "NumberOfNodes": numberOfNodes,
            }
            resultCommand = self.__executeHostCommand("getJobOutputFiles", commandOptions, host=host)
            if not resultCommand["OK"]:
                return resultCommand

            result = resultCommand["Value"]
            if result["Status"] != 0:
                return S_ERROR("Failed to get job output files: %s" % result["Message"])

            if "OutputTemplate" in result:
                self.outputTemplate = result["OutputTemplate"]
                self.errorTemplate = result["ErrorTemplate"]

            output = result["Jobs"][jobStamp]["Output"]
            error = result["Jobs"][jobStamp]["Error"]
        else:
            output = "%s/%s.out" % (self.batchOutput, jobStamp)
            error = "%s/%s.err" % (self.batchError, jobStamp)

        return S_OK((jobStamp, host, output, error))

    def getJobOutput(self, jobID, localDir=None):
        """Get the specified job standard output and error files. If the localDir is provided,
        the output is returned as file in this directory. Otherwise, the output is returned
        as strings.
        """
        self.log.verbose("Getting output for jobID", jobID)
        result = self._getJobOutputFiles(jobID)
        if not result["OK"]:
            return result

        jobStamp, host, outputFile, errorFile = result["Value"]

        if localDir:
            localOutputFile = "%s/%s.out" % (localDir, jobStamp)
            localErrorFile = "%s/%s.err" % (localDir, jobStamp)
        else:
            localOutputFile = "Memory"
            localErrorFile = "Memory"

        # Take into account the SSHBatch possible SSHHost syntax
        host = host.split("/")[0]

        ssh = SSH(host=host, parameters=self.ceParameters)
        result = ssh.scpCall(30, localOutputFile, outputFile, upload=False)
        if not result["OK"]:
            return result

        result = ssh.scpCall(30, localErrorFile, errorFile, upload=False)
        if not result["OK"]:
            return result

        if localDir:
            output = localOutputFile
            error = localErrorFile
        else:
            output = result["Value"][1]
            error = result["Value"][1]

        return S_OK((output, error))



""" DIRAC FileCatalog Security Manager mix-in class with full access checks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.SecurityManagerBase import SecurityManagerBase


class FullSecurityManager(SecurityManagerBase):
    def getPathPermissions(self, paths, credDict):
        """Get path permissions according to the policy"""

        toGet = dict(zip(paths, [[path] for path in paths]))
        permissions = {}
        failed = {}
        res = self.db.fileManager.getPathPermissions(paths, credDict)
        if not res["OK"]:
            return res
        for path, mode in res["Value"]["Successful"].items():
            for resolvedPath in toGet[path]:
                permissions[resolvedPath] = mode
            toGet.pop(path)
        # Copying items because toGet is changed in the cycle
        for path, resolvedPaths in list(toGet.items()):
            if path == "/":
                for resolvedPath in resolvedPaths:
                    permissions[path] = {"Read": True, "Write": True, "Execute": True}
            if os.path.dirname(path) not in toGet:
                toGet[os.path.dirname(path)] = []
            toGet[os.path.dirname(path)] += resolvedPaths
            toGet.pop(path)
        while toGet:
            res = self.db.dtree.getPathPermissions(list(toGet), credDict)
            if not res["OK"]:
                return res
            for path, mode in res["Value"]["Successful"].items():
                for resolvedPath in toGet[path]:
                    permissions[resolvedPath] = mode
                toGet.pop(path)
            for path, error in res["Value"]["Failed"].items():
                if error != "No such file or directory":
                    for resolvedPath in toGet[path]:
                        failed[resolvedPath] = error
                    toGet.pop(path)
            # Copying items because toGet is changed in the cycle
            for path, resolvedPaths in list(toGet.items()):
                if path == "/":
                    for resolvedPath in resolvedPaths:
                        permissions[path] = {"Read": True, "Write": True, "Execute": True}
                if os.path.dirname(path) not in toGet:
                    toGet[os.path.dirname(path)] = []
                toGet[os.path.dirname(path)] += resolvedPaths
                toGet.pop(path)

        if self.db.globalReadAccess:
            for path in permissions:
                permissions[path]["Read"] = True

        return S_OK({"Successful": permissions, "Failed": failed})



""" Collecting utilities for dictionaries
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from itertools import islice


def breakDictionaryIntoChunks(aDict, chunkSize):
    """This function is a generator of chunks of dictionaries of size chunkSize

    :param dict aDict: the dictionary you want to chunk
    :param int chunkSize: the size of the chunks
    :return: a generator object, that generates chunks of the original dictionary in aDict
    """
    iterDict = iter(aDict)
    for _ in range(0, len(aDict), chunkSize):
        yield {k: aDict[k] for k in islice(iterDict, chunkSize)}



""" Graph is a class providing layouts for the complete plot images including
    titles multiple plots and a legend

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import datetime
import time
import os

from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
from DIRAC.Core.Utilities.Graphs.GraphUtilities import pixelToPoint, evalPrefs, to_timestamp, add_time_to_title
from DIRAC.Core.Utilities.Graphs.GraphData import GraphData
from DIRAC.Core.Utilities.Graphs.Legend import Legend
from DIRAC import gLogger

DEBUG = 0


class Graph(object):
    def __init__(self, *args, **kw):
        super(Graph, self).__init__(*args, **kw)

    def layoutFigure(self, legend):

        prefs = self.prefs

        # Get the main Figure object
        # self.figure = Figure()
        figure = self.figure
        self.canvas = FigureCanvasAgg(figure)

        dpi = prefs["dpi"]
        width = float(prefs["width"])
        height = float(prefs["height"])
        width_inch = width / dpi
        height_inch = height / dpi
        figure.set_size_inches(width_inch, height_inch)
        figure.set_dpi(dpi)
        figure.set_facecolor(prefs.get("background_color", "white"))

        figure_padding = float(prefs["figure_padding"])
        figure_left_padding = float(prefs.get("figure_left_padding", figure_padding))
        figure_right_padding = float(prefs.get("figure_right_padding", figure_padding))
        figure_top_padding = float(prefs.get("figure_top_padding", figure_padding))
        figure_bottom_padding = float(prefs.get("figure_bottom_padding", figure_padding))

        text_size = prefs.get("text_size", 8)
        text_padding = prefs.get("text_padding", 5)

        #######################################
        # Make the graph title

        title = prefs.get("title", "")
        subtitle = ""
        title_size = 0
        title_padding = 0
        if title:
            title_size = prefs.get("title_size", 1.5 * text_size)
            title_padding = float(prefs.get("title_padding", 1.5 * text_padding))
            figure.text(
                0.5,
                1.0 - (title_size + figure_padding) / height,
                title,
                ha="center",
                va="bottom",
                size=pixelToPoint(title_size, dpi),
            )

            subtitle = prefs.get("subtitle", "")
            if subtitle:
                sublines = subtitle.split("\n")
                nsublines = len(sublines)
                subtitle_size = prefs.get("subtitle_size", 1.2 * text_size)
                subtitle_padding = float(prefs.get("subtitle_padding", 1.2 * text_padding))
                top_offset = subtitle_size + subtitle_padding + title_size + figure_padding
                for subline in sublines:
                    figure.text(
                        0.5,
                        1.0 - (top_offset) / height,
                        subline,
                        ha="center",
                        va="bottom",
                        size=pixelToPoint(subtitle_size, dpi),
                        fontstyle="italic",
                    )
                    top_offset += subtitle_size + subtitle_padding

        ########################################
        # Evaluate the plot area dimensions
        graph_width = width - figure_left_padding - figure_right_padding
        graph_height = height - figure_top_padding - figure_bottom_padding
        if title:
            graph_height = graph_height - title_padding - title_size
        if subtitle:
            graph_height = graph_height - nsublines * (subtitle_size + subtitle_padding)
        graph_left = figure_left_padding
        graph_bottom = figure_bottom_padding

        #########################################
        # Make the plot time stamp if requested
        flag = prefs.get("graph_time_stamp", True)
        if flag:
            timeString = "Generated on " + datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S ") + "UTC"
            time_size = prefs["text_size"] * 0.8
            figure.text(
                0.995, 0.005, timeString, ha="right", va="bottom", size=pixelToPoint(time_size, dpi), fontstyle="italic"
            )

        #########################################
        # Make the graph Legend if requested

        legend_flag = prefs["legend"]
        legend_ax = None
        column_width = legend.column_width
        if legend_flag:
            legend_position = prefs["legend_position"]
            # legend_width = float(prefs['legend_width'])
            # legend_height = float(prefs['legend_height'])
            legend_width, legend_height, legend_max_height = legend.getLegendSize()
            legend_padding = float(prefs["legend_padding"])
            if legend_position in ["right", "left"]:
                # One column in case of vertical legend
                legend_width = column_width + legend_padding
                legend_height = min(graph_height, legend_max_height)
                bottom = (height - title_size - title_padding - legend_height) / 2.0 / height
                if legend_position == "right":
                    left = 1.0 - (figure_padding + legend_width) / width
                else:
                    left = figure_padding / width
                    graph_left = graph_left + legend_width
                graph_width = graph_width - legend_width - legend_padding
            elif legend_position == "bottom":
                bottom = figure_padding / height
                left = (width - legend_width) / 2.0 / width
                graph_height = graph_height - legend_height - legend_padding
                graph_bottom = graph_bottom + legend_height + legend_padding

            legend_rect = (left, bottom, legend_width / width, legend_height / height)
            legend_ax = figure.add_axes(legend_rect)

        ###########################################
        # Make the plot spots
        plot_grid = prefs["plot_grid"]
        nx = int(plot_grid.split(":")[0])
        ny = int(plot_grid.split(":")[1])

        plot_axes = []
        for j in range(ny - 1, -1, -1):
            for i in range(nx):
                plot_rect = (
                    (graph_left + graph_width * i / nx) / width,
                    (graph_bottom + graph_height * j / ny) / height,
                    graph_width / nx / width,
                    graph_height / ny / height,
                )

                plot_axes.append(figure.add_axes(plot_rect))

        return legend_ax, plot_axes

    def makeTextGraph(self, text="Empty image"):
        """Make an empty text image"""

        self.figure = Figure()
        figure = self.figure
        self.canvas = FigureCanvasAgg(figure)

        prefs = self.prefs
        dpi = prefs["dpi"]
        width = float(prefs["width"])
        height = float(prefs["height"])
        width_inch = width / dpi
        height_inch = height / dpi
        figure.set_size_inches(width_inch, height_inch)
        figure.set_dpi(dpi)
        figure.set_facecolor("white")

        text_size = prefs.get("text_size", 12)
        figure.text(0.5, 0.5, text, horizontalalignment="center", size=pixelToPoint(text_size, dpi))

    def makeGraph(self, data, *args, **kw):

        start = time.time()

        # Evaluate all the preferences
        self.prefs = evalPrefs(*args, **kw)
        prefs = self.prefs

        if DEBUG:
            print("makeGraph time 1", time.time() - start)
            start = time.time()

        if "text_image" in prefs:
            self.makeTextGraph(str(prefs["text_image"]))
            return

        # Evaluate the number of plots and their requested layout
        metadata = prefs.get("metadata", {})
        plot_grid = prefs.get("plot_grid", "1:1")
        nx = int(plot_grid.split(":")[0])
        ny = int(plot_grid.split(":")[1])
        nPlots = nx * ny
        if nPlots == 1:
            if not isinstance(data, list):
                data = [data]
            if not isinstance(metadata, list):
                metadata = [metadata]
        else:
            if not isinstance(data, list):
                # return S_ERROR('Single data for multiplot graph')
                print("Single data for multiplot graph")
                return
            if not isinstance(metadata, list):
                metaList = []
                for _ in range(nPlots):
                    metaList.append(metadata)
                metadata = metaList

        # Initialize plot data
        graphData = []
        plot_prefs = []
        for i in range(nPlots):
            plot_prefs.append(evalPrefs(prefs, metadata[i]))
            gdata = GraphData(data[i])
            if i == 0:
                plot_type = plot_prefs[i]["plot_type"]
            if "sort_labels" in plot_prefs[i]:
                reverse = plot_prefs[i].get("reverse_labels", False)
                gdata.sortLabels(plot_prefs[i]["sort_labels"], reverse_order=reverse)
            if "limit_labels" in plot_prefs[i]:
                if plot_prefs[i]["limit_labels"] > 0:
                    gdata.truncateLabels(plot_prefs[i]["limit_labels"])
            if "cumulate_data" in plot_prefs[i]:
                gdata.makeCumulativeGraph()
            plot_title = plot_prefs[i].get("plot_title", "")
            if plot_title != "NoTitle":
                begin = ""
                end = ""
                if "starttime" in plot_prefs[i] and "endtime" in plot_prefs[i]:
                    begin = to_timestamp(plot_prefs[i]["starttime"])
                    end = to_timestamp(plot_prefs[i]["endtime"])
                elif gdata.key_type == "time":
                    begin = gdata.min_key
                    end = gdata.max_key
                if begin and end:
                    time_title = add_time_to_title(begin, end)
                    if plot_title:
                        plot_title += ":"
                    plot_prefs[i]["plot_title"] = plot_title + " " + time_title
            graphData.append(gdata)

        # Do not make legend for the plot with non-string keys (except for PieGraphs)
        if not graphData[0].subplots and graphData[0].key_type != "string" and not plot_type == "PieGraph":
            prefs["legend"] = False
        if prefs["legend"] and graphData[0].key_type != "string" and plot_type == "PieGraph":
            graphData[0].initialize(key_type="string")

        legend = Legend(graphData[0], None, prefs)
        self.figure = Figure()

        # Make Water Mark
        image = prefs.get("watermark", None)
        self.drawWaterMark(image)

        legend_ax, plot_axes = self.layoutFigure(legend)

        if DEBUG:
            print("makeGraph time layout", time.time() - start)
            start = time.time()

        # Make plots
        for i in range(nPlots):
            plot_type = plot_prefs[i]["plot_type"]
            try:
                # TODO: Remove when we moved to python3
                exec ("import %s" % plot_type)
            except ImportError:
                print("Trying to use python like import")
                try:
                    exec ("from . import  %s" % plot_type)
                except ImportError as x:
                    print("Failed to import graph type %s: %s" % (plot_type, str(x)))
                    return None

            ax = plot_axes[i]
            plot = eval("%s.%s(graphData[i],ax,plot_prefs[i])" % (plot_type, plot_type))
            plot.draw()

        if DEBUG:
            print("makeGraph time plots", time.time() - start)
            start = time.time()

        # Make legend
        if legend_ax:
            legend.setAxes(legend_ax)
            legend.draw()

        if DEBUG:
            print("makeGraph time legend", time.time() - start)
            start = time.time()
        # return S_OK()

    def drawWaterMark(self, imagePath=None):
        """Make the figure water mark"""

        prefs = self.prefs

        try:
            from PIL import Image, ImageEnhance
        except ImportError:
            return

        if not imagePath:
            if "watermark" in prefs:
                imagePath = os.path.expandvars(os.path.expanduser(prefs["watermark"]))

        if not imagePath:
            return

        try:
            image = Image.open(imagePath)
            enh = ImageEnhance.Contrast(image)
            i = enh.enhance(0.1)
            img_size = i.size
            resize = 1.0
            if prefs["width"] < img_size[0]:
                resize = prefs["width"] / float(img_size[0])
            if prefs["height"] < img_size[1]:
                resize = min(resize, prefs["height"] / float(img_size[1]))
            box = (
                0.5 - img_size[0] / float(prefs["width"]) * resize / 2.0,
                0.5 - img_size[1] / float(prefs["height"]) * resize / 2.0,
                img_size[0] / float(prefs["width"]) * resize,
                img_size[1] / float(prefs["height"]) * resize,
            )
            # print box
            ax_wm = self.figure.add_axes(box)
            ax_wm.imshow(i, origin="lower", aspect="equal", zorder=-10)
            ax_wm.axis("off")
            ax_wm.set_frame_on(False)
            ax_wm.set_clip_on(False)
        except Exception as e:
            print(e)

    def writeGraph(self, fname, fileFormat="PNG"):
        """Write out the resulting graph to a file with fname in a given format"""

        self.canvas.draw()
        if fileFormat.lower() == "png":
            self.canvas.print_png(fname)
        else:
            gLogger.error("File format '%s' is not supported!" % fileFormat)
